from dataclasses import dataclass
from typing import *  # type:ignore


@dataclass
class MenuItems:
    """
    MenuItems data model.

    Attributes:
        name: The name of the menu item.
        description: The description of the menu item.
        price: The price of the menu item.
        category: The category of the menu item.
    """

    name: str
    description: str
    price: float
    category: str

    @staticmethod
    def new_empty() -> "MenuItems":
        """
        Creates a new empty MenuItems object.

        Returns:
            MenuItems: A new empty MenuItems object.
        """
        return MenuItems(
            name="",
            description="",
            price=0.0,
            category="",
        )


# initial data
MENUITEMSET: list[MenuItems] = [
    MenuItems(
        name="Hamburger",
        description="A classic hamburger with lettuce, tomato, and onions",
        price=4.99,
        category="Burgers",
    ),
    MenuItems(
        name="Cheeseburger",
        description="A classic cheeseburger with lettuce, tomato, onions and cheese",
        price=5.99,
        category="Burgers",
    ),
    MenuItems(
        name="Fries",
        description="A side of crispy fries",
        price=2.99,
        category="Sides",
    ),
    MenuItems(
        name="Soda",
        description="A refreshing soda",
        price=1.99,
        category="Drinks",
    ),
    MenuItems(
        name="Salad",
        description="A fresh salad",
        price=4.99,
        category="Salads",
    ),
    MenuItems(
        name="Ice Cream",
        description="A sweet treat",
        price=3.99,
        category="Desserts",
    ),
]



import functools
import io
from typing import *  # type: ignore

import rio


async def update_and_apply_theme(
    session: rio.Session,
    theme_replacements: dict[str, Any],
) -> None:
    """
    Overrides the session's theme with the given one, and makes sure to update
    all components so they use the new theme.
    """

    # Build the new theme
    updated_theme = session.theme.replace(**theme_replacements)

    # The theme theme itself can preserve some values. For example, the primary
    # color is encoded in the heading text style, so just replacing the primary
    # palette won't make it go away.
    #
    # When passing the values, make sure to only change those that have been
    # changed relative to the default,
    #
    # Take care of that.
    new_theme = rio.Theme.from_color(
        primary_color=updated_theme.primary_color,
        secondary_color=updated_theme.secondary_color,
        background_color=updated_theme.background_color,
        neutral_color=updated_theme.neutral_color,
        hud_color=updated_theme.hud_color,
        disabled_color=updated_theme.disabled_color,
        success_color=updated_theme.success_color,
        warning_color=updated_theme.warning_color,
        danger_color=updated_theme.danger_color,
        corner_radius_small=updated_theme.corner_radius_small,
        corner_radius_medium=updated_theme.corner_radius_medium,
        corner_radius_large=updated_theme.corner_radius_large,
        color_headings="auto",  # TODO: How to decide this?
        light=updated_theme.background_color.perceived_brightness > 0.5,
    )

    # Apply the theme
    await session._apply_theme(new_theme)

    # The application itself isn't enough, because some components will have
    # read theme values and used them to set e.g. their corner radii. Dirty
    # every component to force a full rebuild.
    for component in session._weak_components_by_id.values():
        session._register_dirty_component(
            component,
            include_children_recursively=False,
        )

    # Refresh
    await session._refresh()


def get_source_for_theme(theme: rio.Theme, *, create_theme_pair: bool) -> str:
    """
    Given a theme, returns a string that can be used to recreate it.
    """
    # Find all values that differ from the defaults
    default_theme = rio.Theme.from_color()
    changed_kwargs = {}

    if theme.primary_color != default_theme.primary_color:
        changed_kwargs["primary_color"] = theme.primary_color

    if theme.secondary_color != default_theme.secondary_color:
        changed_kwargs["secondary_color"] = theme.secondary_color

    if theme.background_color != default_theme.background_color:
        changed_kwargs["background_color"] = theme.background_color

    if theme.neutral_color != default_theme.neutral_color:
        changed_kwargs["neutral_color"] = theme.neutral_color

    if theme.hud_color != default_theme.hud_color:
        changed_kwargs["hud_color"] = theme.hud_color

    if theme.disabled_color != default_theme.disabled_color:
        changed_kwargs["disabled_color"] = theme.disabled_color

    if theme.success_color != default_theme.success_color:
        changed_kwargs["success_color"] = theme.success_color

    if theme.warning_color != default_theme.warning_color:
        changed_kwargs["warning_color"] = theme.warning_color

    if theme.danger_color != default_theme.danger_color:
        changed_kwargs["danger_color"] = theme.danger_color

    if theme.corner_radius_small != default_theme.corner_radius_small:
        changed_kwargs["corner_radius_small"] = round(theme.corner_radius_small, 1)

    if theme.corner_radius_medium != default_theme.corner_radius_medium:
        changed_kwargs["corner_radius_medium"] = round(theme.corner_radius_medium, 1)

    if theme.corner_radius_large != default_theme.corner_radius_large:
        changed_kwargs["corner_radius_large"] = round(theme.corner_radius_large, 1)

    if theme.neutral_color.perceived_brightness < 0.5:
        changed_kwargs["light"] = False

    # Build the source
    theme_or_themes = "themes" if create_theme_pair else "theme"
    result = io.StringIO()
    result.write(f"# Create the {theme_or_themes}\n")

    if create_theme_pair:
        result.write("themes = rio.Theme.pair_from_color(")
    else:
        result.write("theme = rio.Theme.from_color(")

    if changed_kwargs:
        result.write("\n")

        for key, value in changed_kwargs.items():
            result.write(f"    {key}=")

            if isinstance(value, rio.Color):
                hex_value = value.hex
                if len(hex_value) == 8 and hex_value.endswith("ff"):
                    hex_value = hex_value[:-2]

                result.write(f"rio.Color.from_hex({hex_value!r})")
            elif isinstance(value, (int, float)):
                result.write(repr(value))
            else:
                raise NotImplementedError(f"Unsupported type: {type(value)}")

            result.write(",\n")

    result.write(")\n")
    result.write("\n")
    result.write(f"# And apply {'them' if create_theme_pair else 'it'} to your app\n")
    result.write("app = rio.App(\n")
    result.write("    ...\n")
    result.write(f"    theme={theme_or_themes},\n")
    result.write("    ...\n")
    result.write(")")

    # Done
    return result.getvalue()


class PalettePicker(rio.Component):  #
    shared_open_key: str

    palette_nicename: str
    palette_attribute_name: str

    is_colorful_palette: bool

    pick_opacity: bool = False

    round_top: bool = False
    round_bottom: bool = False

    @property
    def palette(self) -> rio.Palette:
        return getattr(self.session.theme, self.palette_attribute_name)

    async def _on_color_change(self, event: rio.ColorChangeEvent) -> None:
        await update_and_apply_theme(
            self.session,
            {
                self.palette_attribute_name: rio.Palette._from_color(
                    event.color,
                    colorful=self.is_colorful_palette,
                )
            },
        )

    def _on_press(self, event: rio.PressEvent) -> None:
        # Toggle the popup
        if self.shared_open_key == self.palette_nicename:
            self.shared_open_key = ""
        else:
            self.shared_open_key = self.palette_nicename

    def build(self) -> rio.Component:
        palette = self.palette

        top_radius = self.session.theme.corner_radius_medium if self.round_top else 0
        bottom_radius = (
            self.session.theme.corner_radius_medium if self.round_bottom else 0
        )

        return rio.Popup(
            anchor=rio.MouseEventListener(
                rio.Rectangle(
                    content=rio.Column(
                        rio.Text(
                            self.palette_nicename,
                            style=rio.TextStyle(
                                # font_size=self.session.theme.heading3_style.font_size,
                                fill=palette.foreground,
                            ),
                            selectable=False,
                            justify="left",
                        ),
                        rio.Text(
                            f"#{palette.background.hex}",
                            style=rio.TextStyle(
                                font_size=1,
                                fill=palette.foreground.replace(opacity=0.5),
                            ),
                            justify="left",
                        ),
                        spacing=0.2,
                        margin_x=1,
                        margin_y=0.8,
                    ),
                    fill=palette.background,
                    corner_radius=(
                        top_radius,
                        top_radius,
                        bottom_radius,
                        bottom_radius,
                    ),
                    ripple=True,
                    transition_time=0.15,
                ),
                on_press=self._on_press,
            ),
            content=rio.Column(
                rio.Text(
                    f"{self.palette_nicename} Color",
                    style="heading3",
                ),
                rio.ColorPicker(
                    color=palette.background,
                    pick_opacity=self.pick_opacity,
                    on_change=self._on_color_change,
                    width=18,
                    height=16,
                ),
                spacing=0.8,
                margin=1,
            ),
            is_open=self.shared_open_key == self.palette_nicename,
            color="hud",
            direction="left",
            gap=1,
        )


class ThemePickerPage(rio.Component):
    shared_open_key: str = ""

    theme_variants_are_initialized: bool = False
    create_light_theme: bool = True
    create_dark_theme: bool = False

    @rio.event.on_populate
    async def _on_populate(self) -> None:
        if self.theme_variants_are_initialized:
            return

        current_theme_is_light = (
            self.session.theme.background_color.perceived_brightness > 0.5
        )

        self.create_light_theme = current_theme_is_light
        self.create_dark_theme = not current_theme_is_light

    async def _on_radius_change(
        self,
        radius_name: str,
        event: rio.SliderChangeEvent,
    ) -> None:
        await update_and_apply_theme(
            self.session,
            {
                radius_name: event.value,
            },
        )

    def _toggle_create_light_theme(self) -> None:
        self.create_light_theme = not self.create_light_theme

        if not self.create_light_theme and not self.create_dark_theme:
            self.create_dark_theme = True

    def _toggle_create_dark_theme(self) -> None:
        self.create_dark_theme = not self.create_dark_theme

        if not self.create_light_theme and not self.create_dark_theme:
            self.create_light_theme = True

    def build(self) -> rio.Component:
        # Prepare the radius sliders
        slider_min = 0
        slider_max = 4
        radius_sliders = rio.Grid(
            (
                rio.Text("Small"),
                rio.Slider(
                    value=self.session.theme.corner_radius_small,
                    minimum=slider_min,
                    maximum=slider_max,
                    width="grow",
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_small",
                    ),
                ),
            ),
            (
                rio.Text("Medium"),
                rio.Slider(
                    value=self.session.theme.corner_radius_medium,
                    minimum=slider_min,
                    maximum=slider_max,
                    width="grow",
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_medium",
                    ),
                ),
            ),
            (
                rio.Text("Large"),
                rio.Slider(
                    value=self.session.theme.corner_radius_large,
                    minimum=slider_min,
                    maximum=slider_max,
                    width="grow",
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_large",
                    ),
                ),
            ),
        )

        # Combine everything
        return rio.ScrollContainer(
            content=rio.Column(
                # Main Colors
                # rio.Text(
                #     "Theme Colors",
                #     style="heading3",
                #     margin_bottom=1,
                #     justify='left',
                # ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Primary",
                    palette_attribute_name="primary_palette",
                    is_colorful_palette=False,
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Secondary",
                    palette_attribute_name="secondary_palette",
                    is_colorful_palette=False,
                    round_bottom=True,
                ),
                # Neutral Colors
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Background",
                    palette_attribute_name="background_palette",
                    is_colorful_palette=False,
                    margin_top=1,
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Neutral",
                    palette_attribute_name="neutral_palette",
                    is_colorful_palette=False,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="HUD",
                    palette_attribute_name="hud_palette",
                    is_colorful_palette=False,
                    pick_opacity=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Disabled",
                    palette_attribute_name="disabled_palette",
                    is_colorful_palette=False,
                    round_bottom=True,
                ),
                # Semantic Colors
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Success",
                    palette_attribute_name="success_palette",
                    is_colorful_palette=True,
                    margin_top=1,
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Warning",
                    palette_attribute_name="warning_palette",
                    is_colorful_palette=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Danger",
                    palette_attribute_name="danger_palette",
                    is_colorful_palette=True,
                    round_bottom=True,
                ),
                # Corner radii
                rio.Text(
                    "Corner Radii",
                    style="heading3",
                    margin_top=1,
                    margin_bottom=1,
                    justify="left",
                ),
                radius_sliders,
                # Theme Variants
                rio.Text(
                    "Theme Variants",
                    style="heading3",
                    margin_top=1,
                    margin_bottom=1,
                    justify="left",
                ),
                rio.Row(
                    rio.Spacer(),
                    rio.IconButton(
                        "material/light-mode",
                        style="minor" if self.create_light_theme else "plain",
                        on_press=self._toggle_create_light_theme,
                    ),
                    rio.Spacer(),
                    rio.IconButton(
                        "material/dark-mode",
                        style="minor" if self.create_dark_theme else "plain",
                        on_press=self._toggle_create_dark_theme,
                    ),
                    rio.Spacer(),
                ),
                # Code Sample
                rio.Text(
                    "Code Sample",
                    style="heading3",
                    margin_top=1,
                    # margin_bottom=1,  Not used for now, since markdown has an oddly large margin anyway
                    justify="left",
                ),
                rio.Markdown(
                    f"""
```python
{get_source_for_theme(self.session.theme, create_theme_pair=self.create_light_theme and self.create_dark_theme)}
```
                    """,
                ),
                margin=1,
                align_y=0,
            ),
            scroll_x="never",
        )



from __future__ import annotations

import enum
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "MouseEventListener",
    "MouseButton",
    "PressEvent",
    "MouseDownEvent",
    "MouseUpEvent",
    "MouseMoveEvent",
    "MouseEnterEvent",
    "MouseLeaveEvent",
    "DragStartEvent",
    "DragMoveEvent",
    "DragEndEvent",
]


@final
class MouseButton(enum.Enum):
    LEFT = "left"
    MIDDLE = "middle"
    RIGHT = "right"


@dataclass
class _MouseUpDownEvent:
    button: MouseButton
    x: float
    y: float


@final
class PressEvent(_MouseUpDownEvent):
    pass


@final
class MouseDownEvent(_MouseUpDownEvent):
    pass


@final
class MouseUpEvent(_MouseUpDownEvent):
    pass


@dataclass
class _MousePositionedEvent:
    x: float
    y: float


@final
class MouseMoveEvent(_MousePositionedEvent):
    pass


@final
class MouseEnterEvent(_MousePositionedEvent):
    pass


@final
class MouseLeaveEvent(_MousePositionedEvent):
    pass


@dataclass
class _DragEvent:
    button: MouseButton
    x: float
    y: float
    component: rio.Component


@final
class DragStartEvent(_DragEvent):
    pass


@final
class DragMoveEvent(_DragEvent):
    pass


@final
class DragEndEvent(_DragEvent):
    pass


@final
class MouseEventListener(FundamentalComponent):
    """
    Allows you to listen for mouse events on a component.

    `MouseEventListeners` take a single child component and display it. They
    then listen for any mouse activity on the child component and report it
    through their event handlers.

    Attributes:
        content: The child component to display.

        on_press: Similar to `on_mouse_up`, but performs additional subtle
            checks, such as that the left mouse button was pressed.

        on_mouse_down: Triggered when a mouse button is pressed down while
            the mouse is placed over the child component.

        on_mouse_up: Triggered when a mouse button is released while the
            mouse is placed over the child component.

        on_mouse_move: Triggered when the mouse is moved while located over
            the child component.

        on_mouse_enter: Triggered when the mouse previously was not located
            over the child component, but now is.

        on_mouse_leave: Triggered when the mouse previously was located over
            the child component, but now is not.
    """

    content: rio.Component
    _: KW_ONLY
    on_press: rio.EventHandler[PressEvent] = None
    on_mouse_down: rio.EventHandler[MouseDownEvent] = None
    on_mouse_up: rio.EventHandler[MouseUpEvent] = None
    on_mouse_move: rio.EventHandler[MouseMoveEvent] = None
    on_mouse_enter: rio.EventHandler[MouseEnterEvent] = None
    on_mouse_leave: rio.EventHandler[MouseLeaveEvent] = None
    on_drag_start: rio.EventHandler[DragStartEvent] = None
    on_drag_move: rio.EventHandler[DragMoveEvent] = None
    on_drag_end: rio.EventHandler[DragEndEvent] = None

    def _custom_serialize(self) -> JsonDoc:
        return {
            "reportPress": self.on_press is not None,
            "reportMouseDown": self.on_mouse_down is not None,
            "reportMouseUp": self.on_mouse_up is not None,
            "reportMouseMove": self.on_mouse_move is not None,
            "reportMouseEnter": self.on_mouse_enter is not None,
            "reportMouseLeave": self.on_mouse_leave is not None,
            "reportDragStart": self.on_drag_start is not None,
            "reportDragMove": self.on_drag_move is not None,
            "reportDragEnd": self.on_drag_end is not None,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        msg_type = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Dispatch the correct event
        if msg_type == "press":
            await self.call_event_handler(
                self.on_press,
                PressEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseDown":
            await self.call_event_handler(
                self.on_mouse_down,
                MouseDownEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseUp":
            await self.call_event_handler(
                self.on_mouse_up,
                MouseUpEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseMove":
            await self.call_event_handler(
                self.on_mouse_move,
                MouseMoveEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "mouseEnter":
            await self.call_event_handler(
                self.on_mouse_enter,
                MouseEnterEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "mouseLeave":
            await self.call_event_handler(
                self.on_mouse_leave,
                MouseLeaveEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "dragStart":
            await self.call_event_handler(
                self.on_drag_start,
                DragStartEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[msg["component"]],
                ),
            )

        elif msg_type == "dragMove":
            await self.call_event_handler(
                self.on_drag_move,
                DragMoveEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[msg["component"]],
                ),
            )

        elif msg_type == "dragEnd":
            await self.call_event_handler(
                self.on_drag_end,
                DragEndEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[msg["component"]],
                ),
            )

        else:
            raise ValueError(f"{__class__.__name__} encountered unknown message: {msg}")

        # Refresh the session
        await self.session._refresh()


MouseEventListener._unique_id = "MouseEventListener-builtin"



from typing import *  # type: ignore

import rio.components.class_container
import rio.components.debugger_connector

from . import (
    deploy_page,
    docs_page,
    icons_page,
    project_page,
    theme_picker_page,
    tree_page,
)


class ClientSideDebugger(rio.Component):
    selected_page: (
        Literal[
            "project",
            "tree",
            "docs",
            "deploy",
        ]
        | None
    ) = None

    def get_selected_page(self) -> rio.Component | None:
        PAGE_WIDTH = 22

        # Nothing selected
        if self.selected_page is None:
            return None

        # Project
        if self.selected_page == "project":
            return project_page.ProjectPage(
                width=PAGE_WIDTH,
            )

        # Tree
        if self.selected_page == "tree":
            return tree_page.TreePage(
                width=PAGE_WIDTH,
            )

        # Icons
        if self.selected_page == "icons":
            return icons_page.IconsPage(
                # This page contains wide source code. Constant changes to the
                # size would cause unsightly resizes.
                width=PAGE_WIDTH + 15
            )

        # Theme
        if self.selected_page == "theme":
            return theme_picker_page.ThemePickerPage(
                width=35,
            )

        # Docs
        if self.selected_page == "docs":
            return docs_page.DocsPage(
                width=PAGE_WIDTH,
            )

        # Deploy
        if self.selected_page == "deploy":
            return deploy_page.DeployPage(
                width=PAGE_WIDTH,
            )

        # Anything else / TODO
        return rio.Text(
            f"TODO: {self.selected_page}",
            margin=2,
            width=PAGE_WIDTH,
        )

    def build(self) -> rio.Component:
        return rio.Row(
            # Big fat line to separate the debugger from the rest of the page
            rio.Rectangle(
                width=0.3,
                fill=self.session.theme.primary_palette.background,
            ),
            # Currently active page
            rio.components.class_container.ClassContainer(
                rio.Switcher(self.get_selected_page()),
                classes=["rio-switcheroo-neutral", "rio-debugger-background"],
            ),
            # Navigation
            rio.Column(
                rio.SwitcherBar(
                    names=[
                        "Project",
                        "Tree",
                        "Icons",
                        "Theme",
                        "Docs",
                        "Deploy",
                    ],
                    icons=[
                        "rio/logo",
                        "material/view-quilt",
                        "material/emoji-people",
                        "material/palette",
                        "material/library-books",
                        "material/rocket-launch",
                    ],
                    values=[
                        "project",
                        "tree",
                        "icons",
                        "theme",
                        "docs",
                        "deploy",
                    ],
                    allow_none=True,
                    orientation="vertical",
                    spacing=2,
                    color="primary",
                    selected_value=self.bind().selected_page,
                    margin=0.2,
                ),
                rio.Spacer(),
                rio.components.debugger_connector.DebuggerConnector(),
            ),
        )



import sys

if sys.platform == "win32":
    from . import (
        IocpProactor_accept_locals_accept_coro,
        ProactorBasePipeTransport_call_connection_lost,
    )

    del IocpProactor_accept_locals_accept_coro
    del ProactorBasePipeTransport_call_connection_lost



from __future__ import annotations

from dataclasses import KW_ONLY
from pathlib import Path
from typing import Literal, final

from uniserde import JsonDoc

import rio

from .. import color, fills, icon_registry
from .fundamental_component import FundamentalComponent

__all__ = [
    "Icon",
]


@final
class Icon(FundamentalComponent):
    """
    # Icon

    Displays one of many pre-bundled icons.

    Icons are a great way to add polish to your app. A good icon can help your
    users understand your app and immediately recognize what a component does.

    Rio includes hundreds of free icons, allowing you to easily add them to your
    app without having to find or create your own. The `Icon` component displays
    one of these icons.

    Note that unlike most components in Rio, `Icon` does not have a `natural`
    size, they can be easily be scaled to fit any size. Therefore it defaults to
    a width and height of 1.3, which is a great size when mixing icons with
    text.

    Icon names are in the format `set_name/icon_name:variant`. Rio already ships
    with the `material` icon set, which contains icons in the style of Google's
    Material Design. You can browse all available icons in Rio's debugger
    sidebar. (The debugger sidebar is visible if you run your project using `rio
    run`.)

    The set name and variant can be omitted. If no set name is specified, it
    defaults to `material`. If no variant is specified, the default version of
    the icon, i.e. no variant, is used.


    ## Attributes:

    `icon`: The name of the icon to display, in the format
        `icon-set/name:variant`. You can browse all available icons in Rio's
        debugger sidebar.

    `fill`: The color scheme of the icon. The text color is used if no fill is
        specified.


    ## Example:

    This minimal example will simply display an Icon:

    ```python
    rio.Icon("material/castle")
    ```

    You can also specify the fill, width and height of the icon. This example
    will display a simple `Icon` named "castle" from the `material` icon set,
    filled in red color and of size 2.5 x 2.5:

    ```python
    rio.Icon(
        "material/castle",
        fill=rio.Color.from_hex("ff0000"),
        height=2.5,
        width=2.5,
    )
    ```
    """

    icon: str
    _: KW_ONLY
    fill: rio.FillLike | rio.ColorSet | Literal["dim"]

    @staticmethod
    def _get_registry() -> icon_registry.IconRegistry:
        return icon_registry.IconRegistry.get_singleton()

    @staticmethod
    def register_icon_set(
        set_name: str,
        icon_set_archive_path: Path,
    ) -> None:
        """
        Add an icon set to the global registry. This allows the icons to be
        accessed as `icon_name`, `set_name/icon_name` or
        `set_name/icon_name:variant`.

        There must not already be a set with the given name.

        The icon set is a `.tar.xz` compressed archive and must contain exactly
        one directory, which must be named identically to the icon set. Files
        located in the root of that directory can be accessed as
        `set_name/icon_name`. Files located in a subdirectory can be accessed as
        `set_name/icon_name:variant`.

        For SVG files to work as icons...

        - They must have a `viewBox` attribute, but no height or width
        - They must contain exactly one XML root node: `<svg>...</svg>`. This
          also goes for comments!
        - Rio colors paths by assigning a `fill` to the SVG root. This only
          works as long as the SVG paths don't have a `<style>` assigned
          already.

        Args:
            set_name: The name of the new icon set. This will be used to access
                the icons.

            icon_set_archive_path: The path to the `.tar.xz` archive containing
            the icon
                set.
        """
        registry = Icon._get_registry()

        if set_name in registry.icon_set_archives:
            raise ValueError(f"There is already an icon set named `{set_name}`")

        registry.icon_set_archives[set_name] = icon_set_archive_path

    @staticmethod
    def register_single_icon(
        icon_source: Path,
        set_name: str,
        icon_name: str,
        variant_name: str | None = None,
    ) -> None:
        """
        Add a single icon to the global registry. This allows the icon to be
        accessed as `icon_name`, `set_name/icon_name` or
        `set_name/icon_name:variant`.

        `icon_source` needs to be the path to a single SVG file. For SVG files
        to work as icons...

        - They must have a `viewBox` attribute, but no height or width
        - They must contain exactly one XML root node: `<svg>...</svg>`. This
          also goes for comments!
        - Rio colors paths by assigning a `fill` to the SVG root. This only
          works as long as the SVG paths don't have a `<style>` assigned
          already.

        Args:
            icon_source: The path to the SVG file containing the icon.

            set_name: The name of the new icon set. This will be used to access
                the icons.

            icon_name: The name of the icon. This will be used to access the
                icon.

            variant_name: The name of the variant. This will be used to access
                the icon. If not specified, the default variant will be used.
        """

        # Try to load the icon
        with open(icon_source) as f:
            svg_source = f.read()

        # Add it to the icon registry's cache
        if variant_name is None:
            name = f"{set_name}/{icon_name}"
        else:
            name = f"{set_name}/{icon_name}:{variant_name}"

        registry = Icon._get_registry()
        registry.cached_icons[name] = svg_source

    def __init__(
        self,
        icon: str,
        *,
        fill: rio.FillLike | rio.ColorSet | Literal["dim"] = "keep",
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["grow"] = 1.3,
        height: float | Literal["grow"] = 1.3,
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.fill = fill

    def __post_init__(self):
        # Verify that the icon exists. We want to crash now, not during the next
        # refresh.
        registry = Icon._get_registry()
        registry.get_icon_svg(self.icon)

    def _custom_serialize(self) -> JsonDoc:
        registry = Icon._get_registry()
        svg_source = registry.get_icon_svg(self.icon)

        # Serialize the fill. This isn't automatically handled because it's a
        # Union.
        if isinstance(self.fill, fills.Fill):
            fill = self.fill._serialize(self.session)
        elif isinstance(self.fill, color.Color):
            fill = self.fill.rgba
        else:
            assert isinstance(self.fill, str), f"Unsupported fill type: {self.fill}"
            fill = self.fill

        # Serialize
        return {
            "svgSource": svg_source,
            "fill": fill,
        }


Icon._unique_id = "Icon-builtin"



from __future__ import annotations

from typing import final

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["ScrollTarget"]


@final
class ScrollTarget(FundamentalComponent):
    """
    # ScrollTarget

    Allows browsers to scroll to a specific component via URL fragment.

    `ScrollTarget` is a container which can be referenced by a URL fragment,
    which allows browsers to scroll to it when the page is loaded. For example,
    if your website contains a `ScrollTarget` with the ID `"my-section"`, then a
    browser visiting `https://your.website/#my-section` will immediately scroll
    it into view.


    ## Attributes:

    `id`: The ID of the `ScrollTarget`. This must be unique among all
            `ScrollTarget`s on the page.

    `content`: The child component to display inside the `ScrollTarget`.


    ## Example:

    A minimal example of `ScrollTarget` displaying an icon:

    #TODO: Better example
    ```python
    rio.ScrollTarget(
        id="my-section",
        content=rio.Icon("material/castle", width=50, height=50),
    )
    ```
    """

    id: str
    content: rio.Component | None = None


ScrollTarget._unique_id = "ScrollTarget-builtin"



from typing import final

from .fundamental_component import FundamentalComponent

__all__ = ["Html"]


@final
class Html(FundamentalComponent):
    """
    # Html

    Renders HTML.

    `Html` allows you to embed HTML in your app. It takes HTML code as input and
    renders it. If you want to embed an entire website, use the `Website`
    component instead.


    ## Attributes:

    `html`: The HTML to render.

    ## Example:

    This minimal example will render a simple HTML code:
    TODO: Add example
    """

    html: str


Html._unique_id = "Html-builtin"



from __future__ import annotations

import collections
import dataclasses
import inspect
import textwrap
import typing
from dataclasses import is_dataclass
from typing import *  # type: ignore

import introspection.typing
import revel
from uniserde import Jsonable

from rio import inspection

from . import models


def split_docstring_into_sections(docstring: str) -> tuple[str, dict[str, str]]:
    """
    Splits the given docstring into sections separated by markdown headings. The
    result is the part of the string before the first heading, and a dictionary
    mapping the heading names to the text of the sections ("the summary").

    If the docstring starts with a heading it will not create a separate section
    and the text of that section will be considered part of the summary.
    """
    # Drop the title if it exists
    if docstring.startswith("#"):
        docstring = docstring.split("\n", 1)[1]

    # Split the docstring into sections
    sections: dict[str, list[str]] = {}
    details: list[str] = []
    current_section: list[str] = details

    # Process the individual lines
    for line in docstring.splitlines():
        if line.startswith("#"):
            section_name = line.strip()
            current_section = sections.setdefault(section_name, [])
        else:
            current_section.append(line)

    # Post-process the sections
    def postprocess(section: list[str]) -> str:
        return "\n".join(section).strip()

    return postprocess(details), {
        name: postprocess(section) for name, section in sections.items()
    }


def parse_key_value_section(section_string: str) -> dict[str, str]:
    """
    Some docstring sections are formatted as a list of key-value pairs, such as
    this:

    ```
    key: A description of the key.
    `key_in_code`: A description of the key in code.
    ```

    This function splits such a section into a dictionary of key-value pairs.
    """
    result_lines: dict[str, list[str]] = {}
    current_value: list[str] = []

    # Process the lines individually
    for raw_line in section_string.splitlines():
        # Strip the line and calculate the indentation
        strip_line = raw_line.lstrip()
        indent = len(raw_line) - len(strip_line)
        strip_line = strip_line.rstrip()

        # Continuation of the previous value
        if indent > 0 or not strip_line:
            current_value.append(strip_line)
            continue

        # New value
        parts = strip_line.split(":", 1)

        if len(parts) == 1:
            revel.warning(
                f'Expected a new mapping (key: value), but got "{strip_line}"'
            )
            current_value.append(strip_line)
            continue

        key, value = parts
        key = key.strip().strip("`").strip()
        value = value.strip()
        current_value = [value]
        result_lines[key] = current_value

    # Convert the section values from lists to strings
    result: dict[str, str] = {}

    for key, lines in result_lines.items():
        result[key] = " ".join(lines).strip()

    return result


def parse_details(details: str) -> tuple[str | None, str | None]:
    """
    Given the details part of a docstring, split it into summary and details.
    Either value may be Nonne if they are not present in the original string.
    """
    details = details.strip()

    # Split into summary & details
    lines = details.split("\n")

    short_lines: list[str] = []
    long_lines: list[str] = []
    cur_lines = short_lines

    for raw_line in lines:
        strip_line = raw_line.strip()

        cur_lines.append(raw_line)

        if not strip_line:
            cur_lines = long_lines

    # Join the lines
    short_description = "\n".join(short_lines).strip()
    long_description = "\n".join(long_lines).strip()

    if not short_description:
        short_description = None

    if not long_description:
        long_description = None

    # Done
    return short_description, long_description


def str_type_hint(typ: type) -> str:
    # Make sure the type annotation has been parsed
    assert not isinstance(typ, str), typ

    # Then pretty-string the type
    return introspection.typing.annotation_to_string(typ)


def parse_docstring_basic(
    docstring: str,
) -> tuple[str | None, str | None, dict[str, str]]:
    """
    Parses a docstring into

    - summary
    - details
    - sections
    """

    # Dedent & strip
    docstring = textwrap.dedent(docstring).strip()

    # Split the docstring into sections
    details, sections = split_docstring_into_sections(docstring)

    # Split into summary and details
    summary, details = parse_details(details)

    # Done
    return summary, details, sections


def parse_docstring(
    docstring: str,
    *,
    key_sections: Iterable[str],
) -> tuple[str | None, str | None, dict[str, dict[str, str]]]:
    """
    Parses a docstring into

    - summary
    - details
    - sections

    Any sections listed in `key_sections` will be parsed as key-value pairs and
    returned as sections. Any remaining sections will be re-joined into the
    details.

    Any sections listed in `key_sections` that are not present in the docstring
    will be imputed as empty.
    """
    # Parse the docstring
    summary, details, sections = parse_docstring_basic(docstring)

    if details is None:
        details = ""

    # Find and parse all key-value sections
    key_sections = set(key_sections)
    key_value_sections: dict[str, dict[str, str]] = {}

    for section_name, section_contents in sections.items():
        # Key section?
        normalized_section_name = section_name.lstrip("#").strip().lower()

        if normalized_section_name in key_sections:
            key_value_sections[normalized_section_name] = parse_key_value_section(
                section_contents
            )

        # Text section
        else:
            details += f"\n\n{section_name}\n\n{section_contents}"

    # Don't keep around an empty details section
    details = details.strip()

    if not details:
        details = None

    # Make sure all requested sections are present
    missing_sections = key_sections - key_value_sections.keys()

    for missing_section in missing_sections:
        key_value_sections[missing_section] = {}

    # Done
    return summary, details, key_value_sections


def parse_function(func: Callable[..., Any]) -> models.FunctionDocs:
    """
    Given a function, parse its signature an docstring into a `FunctionDocs`
    object.
    """

    def parse_annotation(annotation: object) -> str | None:
        if annotation is inspect.Parameter.empty:
            return None

        # Recursive types can cause issues. In particular, we have `Jsonable`,
        # which is a recursive type, and `JsonDoc`, which references `Jsonable`.
        # So if a module imports `JsonDoc` but doesn't import `Jsonable`, it's
        # impossible to evaluate the forward reference `"Jsonable"` in that
        # module.
        #
        # A similar problem exists due to dataclasses: Every `Component`
        # subclass receives a constructor based on the type annotations in
        # `Component`, but of course the relevant imports aren't there.
        #
        # As a workaround, we'll make the missing names available in every
        # module.
        extra_globals = collections.ChainMap(
            {"Jsonable": Jsonable},
            vars(typing),  # type: ignore
        )

        annotation = introspection.typing.resolve_forward_refs(
            annotation,  # type: ignore
            func.__module__,
            extra_globals=extra_globals,
            mode="ast",
            treat_name_errors_as_imports=True,
            strict=True,
        )

        return str_type_hint(annotation)  # type: ignore

    # Parse the parameters
    signature = inspect.signature(func)
    parameters: dict[str, models.FunctionParameter] = {}

    for param_name, param in signature.parameters.items():
        if param.default == inspect.Parameter.empty:
            param_default = None
        else:
            param_default = repr(param.default)

        parameters[param_name] = models.FunctionParameter(
            name=param_name,
            type=parse_annotation(param.annotation),
            default=param_default,
            kw_only=param.kind == inspect.Parameter.KEYWORD_ONLY,
            collect_positional=param.kind == inspect.Parameter.VAR_POSITIONAL,
            collect_keyword=param.kind == inspect.Parameter.VAR_KEYWORD,
            description=None,
        )

    # Parse the docstring
    docstring = inspect.getdoc(func)

    if docstring is None:
        summary = None
        details = None
        raises = []
    else:
        summary, details, sections = parse_docstring(
            docstring,
            key_sections=["args", "raises"],
        )

        raw_params = sections["args"]
        raw_raises = sections["raises"]

        # Add any information learned about parameters from the docstring
        for param_name, param_details in raw_params.items():
            try:
                result_param = parameters[param_name]
            except KeyError:
                revel.warning(
                    f"The docstring for function `{func.__name__}` mentions a parameter `{param_name}` that does not exist in the function signature."
                )
                continue

            result_param.description = param_details

        # Add information about raised exceptions
        raises = list(raw_raises.items())

    # Build the result
    return models.FunctionDocs(
        name=func.__name__,
        parameters=list(parameters.values()),
        return_type=parse_annotation(signature.return_annotation),
        synchronous=not inspect.iscoroutinefunction(func),
        summary=summary,
        details=details,
        raises=raises,
    )


def _parse_class_docstring_with_inheritance(
    cls: type,
    *,
    key_sections: Iterable[str],
) -> tuple[str | None, str | None, dict[str, dict[str, str]]]:
    """
    Parses the docstring of a class in the same format as `parse_docstring`, but
    accounts for inheritance: Key-Value sections of all classes are merged, in a
    way that preserves child docs over parent docs.
    """

    # Parse the docstring for this class
    raw_docs = inspect.getdoc(cls)
    key_sections = set(key_sections)
    parsed_docs = parse_docstring(
        "" if raw_docs is None else raw_docs,
        key_sections=key_sections,
    )

    # Get the docstrings for the base classes
    base_docs: list[tuple[str | None, str | None, dict[str, dict[str, str]]]] = []

    for base in cls.__bases__:
        base_docs.append(
            _parse_class_docstring_with_inheritance(
                base,
                key_sections=key_sections,
            )
        )

    # Merge the docstrings
    result_sections: dict[str, dict[str, str]] = {}
    all_in_order = base_docs + [parsed_docs]

    for docs in all_in_order:
        for section_name, section in docs[2].items():
            result_section = result_sections.setdefault(section_name, {})
            result_section.update(section)

    # Done
    return parsed_docs[0], parsed_docs[1], result_sections


def parse_class(cls: type) -> models.ClassDocs:
    """
    Given a class, parse its signature an docstring into a `ClassDocs` object.
    """

    # Parse the functions
    #
    # Make sure to add functions from base classes as well
    functions_by_name: dict[str, models.FunctionDocs] = {}

    def add_functions(cls: type) -> None:
        # Chain to the base classes
        for base in cls.__bases__:
            add_functions(base)

        # Then process this class. This way they override inherited functions.
        for name, func in inspect.getmembers(cls, inspect.isfunction):
            func_docs = parse_function(func)
            functions_by_name[name] = func_docs

    add_functions(cls)
    functions = list(functions_by_name.values())

    # Parse the fields
    fields_by_name: dict[str, models.ClassField] = {}

    for name, typ in inspection.get_resolved_type_annotations(cls).items():
        if typ is dataclasses.KW_ONLY:
            continue

        fields_by_name[name] = models.ClassField(
            name=name,
            type=str_type_hint(typ),
            default=None,
            description=None,
        )

    # Properties are also fields
    for name, _ in inspect.getmembers(cls, inspect.isdatadescriptor):
        if name in fields_by_name:
            continue

    # Is this a dataclass? If so, get the fields from there
    if is_dataclass(cls):
        for dataclass_field in dataclasses.fields(cls):
            doc_field = fields_by_name[dataclass_field.name]

            # Default value?
            if dataclass_field.default is not dataclasses.MISSING:
                doc_field.default = repr(dataclass_field.default)

            # Default factory?
            elif dataclass_field.default_factory is not dataclasses.MISSING:
                doc_field.default = repr(dataclass_field.default_factory())

    # Parse the docstring
    (
        short_description,
        long_description,
        sections,
    ) = _parse_class_docstring_with_inheritance(
        cls,
        key_sections=["attributes"],
    )

    # Add any information learned about fields from the docstring
    raw_attributes = sections["attributes"]

    for field_name, field_details in raw_attributes.items():
        try:
            result_field = fields_by_name[field_name]
        except KeyError:
            revel.warning(
                f"The docstring for class `{cls.__name__}` mentions a field `{field_name}` that does not exist in the class."
            )
            continue

        result_field.description = field_details

    # If `__init__` is missing documentation for any parameters, try to copy the
    # values from matching fields.
    for func_docs in functions:
        if func_docs.name != "__init__":
            continue

        for param in func_docs.parameters:
            if param.description is not None:
                continue

            try:
                field = fields_by_name[param.name]
            except KeyError:
                continue

            param.description = field.description

        break

    # Build the result
    return models.ClassDocs(
        name=cls.__name__,
        attributes=list(fields_by_name.values()),
        functions=functions,
        summary=short_description,
        details=long_description,
    )



from typing import final

from .fundamental_component import FundamentalComponent

__all__ = [
    "ProgressBar",
]


@final
class ProgressBar(FundamentalComponent):
    """
    # ProgressBar

    A progress indicator in the shape of a horizontal bar.

    `ProgressBar` conveys to the user that activity is ongoing. It can either
    display the exact progress as a fraction from 0 to 1, or it can display an
    indeterminate progress animation, which is useful when the exact progress
    isn't known.

    The circular counterpart to this component is the `ProgressCircle`.


    ## Attributes:

    `progress`: The progress to display, as a fraction from 0 to 1. If `None`,
            the progress indicator will be indeterminate.


    ## Example:

    A minimal example displaying a progress bar that is 50% complete.

    ```python
    rio.ProgressBar(progress=0.5)
    ```
    """

    progress: float | None = None


ProgressBar._unique_id = "ProgressBar-builtin"



from .fundamental_component import FundamentalComponent

__all__ = [
    "DebuggerConnector",
]


class DebuggerConnector(FundamentalComponent):
    pass


DebuggerConnector._unique_id = "DebuggerConnector-builtin"



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import field
from typing import Literal, final

from typing_extensions import Self

import rio

from .component import Component

__all__ = ["LabeledColumn"]


@final
class LabeledColumn(Component):
    """
    # LabeledColumn

    A container that lays out its children in a column, with labels for each child.

    `LabeledColumn` is a container that lays out its children in a column, with
    labels for each child. It is similar to `Column`, but allows you to specify
    labels for each child.


    ## Attributes:

    `content`: The components to place inside the container, with labels.


    ## Example:

    This minimal example will show a container with a column layout and labels:

    ```python
    rio.LabeledColumn(
        {
            "First Name": rio.TextInput(),
            "Last Name": rio.TextInput(),
            "Age": rio.TextInput(),
        }
    )
    ```
    """

    _child_list: list[Component] = field(init=False)

    def __init__(
        self,
        content: Mapping[str, rio.Component],
        *,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.content = content

    @property
    def content(self) -> Mapping[str, Component]:
        return self._content

    @content.setter
    def content(self, children: Mapping[str, Component]) -> None:
        self._content = dict(children)
        self._child_list = list(children.values())

    def add(self, label: str, child: rio.Component) -> Self:
        """
        Appends a child component.
        """
        self._content[label] = child
        self._child_list.append(child)

        return self

    def build(self) -> Component:
        rows = [
            [
                rio.Text(label, justify="right"),
                child,
            ]
            for label, child in self.content.items()
        ]

        return rio.Grid(*rows, row_spacing=0.1, column_spacing=0.2)



from dataclasses import dataclass
from typing import *  # type: ignore

import uniserde

from . import parsers


@dataclass
class FunctionParameter(uniserde.Serde):
    name: str
    type: str | None
    default: str | None

    kw_only: bool

    collect_positional: bool
    collect_keyword: bool

    description: str | None


@dataclass
class FunctionDocs(uniserde.Serde):
    name: str
    parameters: list[FunctionParameter]
    return_type: str | None
    synchronous: bool

    summary: str | None
    details: str | None

    raises: list[tuple[str, str]]  # type, description

    @classmethod
    def parse(cls, func: Callable) -> "FunctionDocs":
        return parsers.parse_function(func)


@dataclass
class ClassField(uniserde.Serde):
    name: str
    type: str
    default: str | None

    description: str | None


@dataclass
class ClassDocs(uniserde.Serde):
    name: str
    attributes: list[ClassField]  # name, type
    functions: list[FunctionDocs]

    summary: str | None
    details: str | None

    @classmethod
    def parse(cls, typ: Type) -> "ClassDocs":
        return parsers.parse_class(typ)



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc


class StressClient:
    """
    This is a fake client for Rio apps, which allows to send requests to the
    server to find errors and test the server under load.
    """

    def __init__(
        self,
        send_message: Callable[[JsonDoc], Awaitable[None]],
        receive_message: Callable[[], Awaitable[JsonDoc]],
    ) -> None:
        # Used to talk to the Rio server
        self.send_message = send_message
        self.receive_message = receive_message

        # True, if the

    async def serve(self) -> None:
        raise NotImplementedError



import rio

# <additional-imports>
from .. import data_models

# </additional-imports>


# <component>
class ItemEditor(rio.Component):
    """
    A component for editing a menu item.

    Returns a card component containing the menu item editor. The editor contains fields
    for name, description, price, and category of the menu item. The component also contains
    buttons for saving or canceling the changes.

    Attributes:
        currently_selected_menu_item: The currently selected menu item.
        new_entry: A boolean flag indicating if the menu item is a new entry.
        on_cancel_event: An event handler for the cancel button.
        on_save_event: An event handler for the save button.
    """

    currently_selected_menu_item: data_models.MenuItems
    new_entry: bool = False
    on_cancel_event: rio.EventHandler[[]] = None
    on_save_event: rio.EventHandler[[]] = None

    async def on_press_save_event(self) -> None:
        """
        Asynchronously triggers the 'on_save_event' when the save button is pressed.
        """
        await self.call_event_handler(self.on_save_event)

    async def on_press_cancel_event(self) -> None:
        """
        Asynchronously triggers the 'on_cancel_event' when the cancel button is pressed.
        """
        await self.call_event_handler(self.on_cancel_event)

    def on_change_name(self, ev: rio.TextInputChangeEvent) -> None:
        """
        Changes the name of the currently selected menu item.

        Args:
            ev: The event object that contains the new text.
        """
        self.currently_selected_menu_item.name = ev.text

    def on_change_description(self, ev: rio.TextInputChangeEvent) -> None:
        """
        Changes the description of the currently selected menu item.

        Args:
            ev: The event object that contains the new text.
        """
        self.currently_selected_menu_item.description = ev.text

    def on_change_price(self, ev: rio.NumberInputChangeEvent) -> None:
        """
        Changes the price of the currently selected menu item.

        Args:
            ev: The event object that contains the new price.
        """
        self.currently_selected_menu_item.price = ev.value

    def on_change_category(self, ev: rio.DropdownChangeEvent) -> None:
        """
        Changes the category of the currently selected menu item.

        Args:
            ev: The event object that contains the new category.
        """
        self.currently_selected_menu_item.category = ev.value

    def build(self) -> rio.Component:
        """
        Builds the menu item editor component.

        Returns:
            A card component containing the menu item editor.
            See the approx. layout below:

        ################ Card #################
        # Text                                #
        # TextInput (Name)                    #
        # TextInput (Description)             #
        # NumberInput (Price)                 #
        # Dropdown (Category)                 #
        # Button (Save)   |  Button (Cancel)  #
        #######################################
        """

        if self.new_entry is False:
            text = "Edit Menu Item"
        else:
            text = "Add New Menu Item"

        return rio.Card(
            rio.Column(
                rio.Text(
                    text=text,
                    style="heading2",
                    margin_bottom=1,
                ),
                rio.TextInput(
                    self.currently_selected_menu_item.name,
                    label="Name",
                    on_change=self.on_change_name,
                ),
                rio.TextInput(
                    self.currently_selected_menu_item.description,
                    label="Description",
                    on_change=self.on_change_description,
                ),
                rio.NumberInput(
                    self.currently_selected_menu_item.price,
                    label="Price",
                    suffix_text="$",
                    on_change=self.on_change_price,
                ),
                rio.Dropdown(
                    options=["Burgers", "Desserts", "Drinks", "Salads", "Sides"],
                    label="Category",
                    selected_value=self.currently_selected_menu_item.category,
                    on_change=self.on_change_category,
                ),
                rio.Row(
                    rio.Button("Save", on_press=self.on_press_save_event),
                    rio.Button("Cancel", on_press=self.on_press_cancel_event),
                    spacing=1,
                    align_x=1,
                ),
                spacing=1,
                align_y=0,
                margin=2,
            ),
        )


# </component>



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal, final

from typing_extensions import Self
from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Row",
    "Column",
]


class _LinearContainer(FundamentalComponent):
    children: list[rio.Component]
    spacing: float = 0.0
    proportions: Literal["homogeneous"] | Sequence[float] | None = None

    # Don't let @dataclass generate a constructor
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add(self, child: rio.Component) -> Self:
        """
        Appends a child component.
        """
        self.children.append(child)
        return self

    def _custom_serialize(self) -> JsonDoc:
        return {"proportions": self.proportions}  # type: ignore[variance]


@final
class Row(_LinearContainer):
    """
    # Row

    A container that lays out its children horizontally.

    `Row`s are one of the most common components in Rio. They take any number of
    children and lay them out horizontally, with the first one on the left, the
    second one to its right, and so on. All components in `Row`s occupy the full
    height of their parent.

    The `Row`'s horizontal counterpart is the `Column`. A similar component, but
    stacking its children in the Z direction, is the `Stack`.

    ## Undefined Space

    Like most containers in `rio`, `Row`s always attempt to allocate all
    available space to their children. In the context of a `Row` though, this
    could easily lead to unexpected results. If more space is available than
    needed, should all children grow? Only the first? Should they grow by equal
    amounts, or proportionally?

    To avoid this ambiguity, `Row`s have a concept of *undefined space*. Simply
    put, **not using all available space is considered an error and should be
    avoided.** `Row`s indicate this by highlighting the extra space with
    unmistakable animated sprites.

    Getting rid of undefined space is easy: Depending on what look you're going
    for, either add a `Spacer` somewhere into your `Row`, assign one of the
    components a `"grow"` value as its height, or set the `Row`'s vertical
    alignment.

    ## Attributes:

    `children`: The components to place in this `Row`.

    `spacing`: How much empty space to leave between two adjacent children. No
            spacing is added before the first child or after the last child.

    `proportions`: If set, the children will grow according to these proportions.
        - `homogeneous`: All children will grow equally.
        - A list of floats: Each child will grow according to its proportion.
        - `None`: Allows all child components to expand as much as they need.

    ## Example:

    This minimal example will display a `Row` with two text components:

    ```python
    rio.Row(rio.Text("Hello"), rio.Text("World!"))
    ```

    `Row`s are commonly used to line up multiple components horizontally. In this example, we're
    using an Icon and two Text components in a Row and wrap them in a Card.

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Card(
                content=rio.Row(
                    rio.Icon(icon="material/castle"),
                    rio.Text("Hello"),
                    rio.Text("World!"),
                    spacing=1,
                    # align card content in the center
                    # to avoid undefined space
                    align_x=0.5,
                ),
            )
    ```
    """

    def __init__(
        self,
        *children: rio.Component,
        spacing: float = 0.0,
        proportions: Literal["homogeneous"] | Sequence[float] | None = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.spacing = spacing
        self.proportions = proportions


Row._unique_id = "Row-builtin"


@final
class Column(_LinearContainer):
    """
    # Column

    A container that lays out its children vertically.

    `Column`s are one of the most common components in Rio. They take any number
    of children and lay them out vertically, with the first one at the top, the
    second one below that, and so on. All components in `Column`s occupy the
    full width of their parent.

    The `Column`'s horizontal counterpart is the `Row`. A similar component, but
    stacking its children in the Z direction, is the `Stack`.


    ## Undefined Space

    Like most containers in `rio`, `Column`s always attempt to allocate all
    available space to their children. In the context of a `Column` though, this
    could easily lead to unexpected results. If more space is available than
    needed, should all children grow? Only the first? Should they grow by equal
    amounts, or proportionally?

    To avoid this ambiguity, `Column`s have a concept of *undefined space*.
    Simply put, **not using all available space is considered an error and
    should be avoided.** `Column`s indicate this by highlighting the extra space
    with unmistakable animated sprites.

    Getting rid of undefined space is easy: Depending on what look you're going
    for, either add a `Spacer` somewhere into your `Column`, assign one of the
    components a `"grow"` value as its height, or set the `Column`'s vertical
    alignment.


    ## Attributes:

    `children`: The components to place in this `Column`.

    `spacing`: How much empty space to leave between two adjacent children. No
            spacing is added before the first child or after the last child.

    `proportions`: If set, the children will grow according to these proportions.
        - `homogeneous`: All children will grow equally.
        - A list of floats: Each child will grow according to its proportion.
        - `None`: Allows all child components to expand as much as they need.


    ## Example:

    This minimal example will display a `Column` with two text components:

    ```python
    rio.Column(rio.Text("Hello"), rio.Text("World!"))
    ```

    `Columns`s are commonly used to line up multiple components vertically. In
    this example, we're using an Icon and two Text components in a Column and
    wrap them in a Card.

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Card(
                content=rio.Column(
                    rio.Icon("material/castle"),
                    rio.Text("Hello"),
                    rio.Text("World!"),
                    spacing=1,
                    # Align card content in the center to avoid undefined space
                    align_x=0.5,
                ),
            )
    ```
    """

    def __init__(
        self,
        *children: rio.Component,
        spacing: float = 0.0,
        proportions: Literal["homogeneous"] | Sequence[float] | None = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.spacing = spacing
        self.proportions = proportions


Column._unique_id = "Column-builtin"



from pathlib import Path
from typing import *  # type: ignore

import rio

from ... import common


class ComponentDetails(rio.Component):
    component_id: int

    component_natural_width: float = 0
    component_natural_height: float = 0

    component_allocated_width: float = 0
    component_allocated_height: float = 0

    @rio.event.on_populate
    async def fetch_client_side_details(self) -> None:
        """
        Fetch additional details about the component which are only known
        client-side. Stores them in the class.
        """
        # Remember which component the details are being fetched for
        component_id = self.component_id

        # Fetch the details
        response = await self.session._evaluate_javascript(
            f"""
            let component = componentsById[{component_id}];
            let rect = component.element.getBoundingClientRect();

            return [
                component.naturalWidth,
                component.naturalHeight,
                rect.width / pixelsPerRem,
                rect.height / pixelsPerRem,
            ];
            """
        )

        # If the current component has changed while the values were fetched,
        # don't update the state
        if self.component_id != component_id:
            return

        # Publish the results
        (
            self.component_natural_width,
            self.component_natural_height,
            self.component_allocated_width,
            self.component_allocated_height,
        ) = response

    def _get_effective_margins(self) -> Tuple[float, float, float, float]:
        return (
            common.first_non_null(
                self.margin_left,
                self.margin_x,
                self.margin,
                0,
            ),
            common.first_non_null(
                self.margin_top,
                self.margin_y,
                self.margin,
                0,
            ),
            common.first_non_null(
                self.margin_right,
                self.margin_x,
                self.margin,
                0,
            ),
            common.first_non_null(
                self.margin_bottom,
                self.margin_y,
                self.margin,
                0,
            ),
        )

    def build(self) -> rio.Component:
        # Get the target component
        try:
            target = self.session._weak_components_by_id[self.component_id]

        # In rare cases components can't be found, because they're created
        # client-side (e.g. injected margins). In this case, just don't display
        # anything.
        except KeyError:
            return rio.Spacer(height=0)

        # Create a grid with all the details
        result = rio.Grid(row_spacing=0.5, column_spacing=0.5)
        row_index = 0

        def add_cell(
            text: str,
            column_index: int,
            is_label: bool,
            *,
            width: int = 1,
        ) -> None:
            result.add(
                rio.Text(
                    text,
                    style="dim" if is_label else "text",
                    justify="right" if is_label else "left",
                ),
                row_index,
                column_index,
                width=width,
            )

        # Any values which should be displayed right in the title
        header_accessories = []

        # TODO: Display this only if rio itself (not the app!) is in debug mode
        # if self.session._app_server.debug_mode:
        #     header_accessories.append(rio.Text(f"#{target._id}", style="dim"))

        if target.key is not None:
            header_accessories = [
                rio.Icon("material/key", fill="dim"),
                rio.Text(
                    target.key,
                    style="dim",
                    justify="left",
                ),
            ]

        # Title
        result.add(
            rio.Row(
                rio.Text(
                    type(target).__qualname__,
                    style="heading3",
                ),
                rio.Spacer(),
                *header_accessories,
                margin_bottom=0.2,
                spacing=0.5,
            ),
            row_index,
            0,
            width=4,
        )
        row_index += 1

        # Which file/line was this component instantiated from?
        file, line = target._creator_stackframe_

        try:
            file = file.relative_to(Path.cwd())
        except ValueError:
            pass

        result.add(
            rio.Text(
                f"{file} line {line}",
                style="dim",
                justify="left",
            ),
            row_index,
            0,
            width=4,
        )
        row_index += 1

        # Custom properties
        #
        # Make sure to skip any which already have custom tailored cells
        debug_details = target.get_debug_details()
        for prop_name, prop_value in debug_details.items():
            # Some values have special handling below
            if prop_name in (
                "key",
                "width",
                "height",
                "margin",
                "margin_x",
                "margin_y",
                "margin_left",
                "margin_right",
                "margin_top",
                "margin_bottom",
                "align_x",
                "align_y",
            ):
                continue

            # Display this property
            value_limit = 30
            prop_str = repr(prop_value)

            if len(prop_str) > value_limit:
                prop_str = prop_str[: value_limit - 1] + "…"

            add_cell(prop_name, 0, True)
            add_cell(prop_str, 1, False, width=3)
            row_index += 1

        # Size
        if "width" in debug_details or "height" in debug_details:
            try:
                py_width_str = debug_details["width"]
            except KeyError:
                py_width_str = "-"
            else:
                if isinstance(py_width_str, (int, float)):
                    py_width_str = round(py_width_str, 2)

                py_width_str = repr(py_width_str)

            try:
                py_height_str = debug_details["height"]
            except KeyError:
                py_height_str = "-"
            else:
                if isinstance(py_height_str, (int, float)):
                    py_height_str = round(py_height_str, 2)

                py_height_str = repr(py_height_str)

            # Spacing to separate the table from the rest
            row_index += 1

            # Header
            result.add(rio.Text("width", style="dim", justify="left"), row_index, 1)
            result.add(rio.Text("height", style="dim", justify="left"), row_index, 2)
            row_index += 1

            # The size as specified in Python
            add_cell("python", 0, True)
            add_cell(py_width_str, 1, False)
            add_cell(py_height_str, 2, False)
            row_index += 1

            # The component's natural size
            add_cell("natural", 0, True)
            add_cell(str(round(self.component_natural_width, 2)), 1, False)
            add_cell(str(round(self.component_natural_height, 2)), 2, False)
            row_index += 1

            # The component's allocated size
            add_cell("allocated", 0, True)
            add_cell(str(round(self.component_allocated_width, 2)), 1, False)
            add_cell(str(round(self.component_allocated_height, 2)), 2, False)
            row_index += 1

            # More spacing
            row_index += 1

        # Margins
        (
            margin_left,
            margin_top,
            margin_right,
            margin_bottom,
        ) = self._get_effective_margins()

        single_x_margin = margin_left == margin_right
        single_y_margin = margin_top == margin_bottom

        if single_x_margin and single_y_margin:
            add_cell("margin", 0, True)
            add_cell(str(margin_left), 1, False)
            row_index += 1

        else:
            if single_x_margin:
                add_cell("margin_x", 0, True)
                add_cell(str(margin_left), 1, False)

            else:
                add_cell("margin_left", 0, True)
                add_cell(str(margin_left), 1, False)

                add_cell("margin_right", 2, True)
                add_cell(str(margin_right), 3, False)

            row_index += 1

            if single_y_margin:
                add_cell("margin_y", 0, True)
                add_cell(str(margin_top), 1, False)

            else:
                add_cell("margin_top", 0, True)
                add_cell(str(margin_top), 1, False)

                add_cell("margin_bottom", 2, True)
                add_cell(str(margin_bottom), 3, False)

            row_index += 1

        # Alignment
        if "align_x" in debug_details or "align_y" in debug_details:
            add_cell("align_x", 0, True)
            add_cell(str(debug_details.get("align_x", "-")), 1, False)

            add_cell("align_y", 2, True)
            add_cell(str(debug_details.get("align_y", "-")), 3, False)
            row_index += 1

        # Link to docs
        if type(target)._rio_builtin_:
            docs_url = f"https://rio.dev/docs/{type(target).__name__.lower()}"
            link_color = self.session.theme.secondary_color

            result.add(
                rio.Link(
                    rio.Row(
                        rio.Icon("material/library-books", fill=link_color),
                        rio.Text("Docs", style=rio.TextStyle(fill=link_color)),
                        spacing=0.5,
                        align_x=0,
                    ),
                    docs_url,
                    # TODO: Support icons in links
                    # new_tab=True,  # TODO: Support this
                    margin_top=0.2,
                ),
                row_index,
                0,
                width=2,
            )
            row_index += 1

        # Done
        return result



from __future__ import annotations

from typing import *  # type: ignore

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "NodeInput",
]


@final
class NodeInput(FundamentalComponent):
    name: str
    color: rio.Color

    def __init__(
        self,
        name: str,
        color: rio.Color,
        # Note that the key is required. Connections use the port's key to
        # identify their start and end points.
        key: str,
        *,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
    ):
        # Make sure the building component is a Node
        # TODO

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
        )

        self.name = name
        self.color = color


NodeInput._unique_id = "NodeInput-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Separator",
]


@final
class Separator(FundamentalComponent):
    """
    # Separator

    A line to separate content.

    `Separator` is a horizontal or vertical line that can be used to separate content. It
    can be styled with a color. By default, it is a thin line with a light gray
    color.


    ## Attributes:

    `color`: The color of the `Separator`. If `None`, the color will be
            determined by the theme.


    ## Example:

    A minimal example of `Separator`:

    ```python
    rio.Separator()
    ```

    A separator is commonly used to separate content. For example, to separate
    two buttons:

    ```python
    rio.Row(
        rio.Button("First"),
        rio.Separator(),
        rio.Button("Second"),
        align_x=0.5,  # fix undefined space
        spacing=1,
    )
    ```
    """

    _: KW_ONLY
    color: rio.Color | None = None


Separator._unique_id = "Separator-builtin"



from __future__ import annotations

from typing import *  # type: ignore

import rio

from .. import components as comps


# <component>
class EmptyChatPlaceholder(rio.Component):
    """
    This component is a placeholder, which is displayed to the user if there is
    no chat history yet. It greets the user, contains suggestions for the user
    to start a conversation, as well as a text input for the user to ask
    anything they have in mind.
    """

    # This will be used for the text input to store its result in
    user_message_text: str = ""

    # This event will be triggered when the user sends a message, be it custom
    # or one of the suggestions.
    on_question: rio.EventHandler[str] = None

    async def on_text_input_confirm(self, *_) -> None:
        """
        Called when the text input is confirmed, or the "send" button pressed.
        The function ensures that the input isn't empty. If that's the case the
        message is sent on to the `on_question` event.
        """
        # If the user hasn't typed anything, do nothing
        message_text = self.user_message_text.strip()

        if not message_text:
            return

        # Trigger the `on_question` event
        await self.call_event_handler(self.on_question, message_text)

    def build(self) -> rio.Component:
        return rio.Column(
            # Greet the user
            rio.Text(
                "Chat with AI",
                style=rio.TextStyle(
                    font_size=5,
                    font_weight="bold",
                    fill=rio.LinearGradientFill(
                        (self.session.theme.secondary_color, 0),
                        (self.session.theme.primary_color, 1),
                    ),
                ),
            ),
            # Explain what the app is all about
            rio.Text(
                "Ask a free-form question and AI will help you on your journey",
                margin_top=1,
            ),
            # Give the user an opportunity to enter a custom question
            rio.Row(
                rio.MultiLineTextInput(
                    label="Ask something...",
                    text=self.bind().user_message_text,
                    on_confirm=self.on_text_input_confirm,
                    width="grow",
                    height=5,
                ),
                rio.IconButton(
                    "material/navigate-next",
                    on_press=self.on_text_input_confirm,
                ),
                spacing=1,
                margin_top=1,
            ),
            # And also give suggestions for them to start with
            rio.Text(
                "Try asking",
                style="dim",
                margin_top=3,
            ),
            rio.Row(
                comps.ChatSuggestionCard(
                    "material/restaurant",
                    "Suggest ways to make a dish more delicious",
                    on_press=self.on_question,
                ),
                comps.ChatSuggestionCard(
                    "material/coffee",
                    "What's the best way to store coffee?",
                    on_press=self.on_question,
                ),
                comps.ChatSuggestionCard(
                    "material/co-present",
                    "Help me improve my presentation technique",
                    on_press=self.on_question,
                ),
                comps.ChatSuggestionCard(
                    "material/work",
                    "Draft a job application email for me",
                    on_press=self.on_question,
                ),
                spacing=1,
                margin_top=1,
            ),
        )


# </component>



from __future__ import annotations

from collections.abc import Callable
from typing import *  # type: ignore

from .component import Component
from .fundamental_component import FundamentalComponent
from .scroll_container import ScrollContainer

__all__ = ["HighLevelRootComponent"]


# The HighLevelRootComponent is the top-most element. It's a high-level element
# that is never re-built, which is convenient in many ways:
# 1. Every component except for the root component itself has a valid builder
# 2. The JS code is simpler because the root component can't have an
#    alignment or margin
# 3. Children of non-fundamental components are automatically initialized
#    correctly, so we don't need to duplicate that logic here
class HighLevelRootComponent(Component):
    build_function: Callable[[], Component]
    build_connection_lost_message_function: Callable[[], Component]

    def build(self) -> Component:
        # Spawn a debugger if running in debug mode

        if self.session._app_server.debug_mode:
            # Avoid a circular import
            import rio.debug.client_side_debugger

            debugger = rio.debug.client_side_debugger.ClientSideDebugger()
        else:
            debugger = None

        # User content should automatically scroll if it grows too large, so we
        # wrap it in a ScrollContainer
        user_content = ScrollContainer(self.build_function())

        return FundamentalRootComponent(
            user_content,
            self.build_connection_lost_message_function(),
            debugger=debugger,
        )


class FundamentalRootComponent(FundamentalComponent):
    content: Component
    connection_lost_component: Component
    debugger: Component | None


FundamentalRootComponent._unique_id = "FundamentalRootComponent-builtin"



# <additional-imports>
import functools

import rio

from .. import data_models

# </additional-imports>


# <component>
class ItemList(rio.Component):
    """
    A component for displaying a list of menu items.

    Returns a list of menu items in a ListView component. Each item in the list contains the name,
    description, and a delete button. The delete button triggers an event to delete the item.

    Attributes:
        menu_item_set: The list of menu items to be displayed.
        on_add_new_item_event: An event handler for adding a new item.
        on_delete_item_event: An event handler for deleting an item.
        on_select_item_event: An event handler for selecting an item.
    """

    menu_item_set: list[data_models.MenuItems]
    on_add_new_item_event: rio.EventHandler[[]] = None
    on_delete_item_event: rio.EventHandler[int] = None
    on_select_item_event: rio.EventHandler[data_models.MenuItems] = None

    async def on_press_add_new_item_event(self) -> None:
        """
        Asynchronously triggers the 'add new item' when the list item is pressed.
        """
        await self.call_event_handler(self.on_add_new_item_event)

    async def on_press_delete_item_event(self, idx: int) -> None:
        """
        Asynchronously triggers the 'delete item' when the delete button is pressed.
        The event handler is passed the index of the item to be deleted.

        Args:
            idx: The index of the item to be deleted.
        """
        await self.call_event_handler(self.on_delete_item_event, idx)
        # update the list
        await self.force_refresh()

    async def on_press_select_item_event(self, item: data_models.MenuItems) -> None:
        """
        Asynchronously triggers the 'select item' when an item is selected.
        The event handler is passed the selected item.

        Args:
            item: The selected item.
        """
        await self.call_event_handler(self.on_select_item_event, item)

    def build(self) -> rio.Component:
        """
        Builds the component by returning a ListView component containing the menu items.

        Returns:
            A ListView component containing the menu items.
            See the approx. layout below:

        ############### ListView ###############
        #  + Add new                           #
        #  Hamburger          Button(Delete)   #
        #  Cheese Burger      Button(Delete)   #
        #  Fries              Button(Delete)   #
        #  ...                                 #
        ########################################
        """

        # Store all children in an intermediate list
        list_items = []

        list_items.append(
            rio.SimpleListItem(
                text="Add new",
                secondary_text="Description",
                key="add_new",
                left_child=rio.Icon("material/add"),
                on_press=self.on_press_add_new_item_event,
            )
        )

        for i, item in enumerate(self.menu_item_set):
            list_items.append(
                rio.SimpleListItem(
                    text=item.name,
                    secondary_text=item.description,
                    right_child=rio.Button(
                        rio.Icon("material/delete"),
                        color=self.session.theme.danger_color,
                        # Note the use of functools.partial to pass the
                        # index to the event handler.
                        on_press=functools.partial(self.on_press_delete_item_event, i),
                    ),
                    # Use the name as the key to ensure that the list item
                    # is unique.
                    key=item.name,
                    # Note the use of functools.partial to pass the
                    # item to the event handler.
                    on_press=functools.partial(self.on_press_select_item_event, item),
                )
            )

        # Then unpack the list to pass the children to the ListView
        return rio.ListView(
            *list_items,
            align_y=0,
        )


# </component>



from __future__ import annotations

import asyncio
import contextlib
import functools
import inspect
import io
import json
import logging
import secrets
import time
import traceback
import weakref
from datetime import timedelta
from typing import *  # type: ignore
from xml.etree import ElementTree as ET

import fastapi
import pytz
import timer_dict
import uniserde.case_convert
from PIL import Image
from uniserde import Jsonable

import rio

from . import (
    app,
    assets,
    byte_serving,
    common,
    components,
    debug,
    global_state,
    inspection,
    routing,
    session,
    user_settings_module,
)
from .common import URL
from .components.root_components import HighLevelRootComponent
from .errors import AssetError
from .serialization import serialize_json

try:
    import plotly  # type: ignore[missing-import]
except ImportError:
    plotly = None

__all__ = [
    "AppServer",
]


@functools.lru_cache(maxsize=None)
def _build_sitemap(base_url: rio.URL, app: rio.App) -> str:
    # Find all pages to add
    page_urls = {
        rio.URL(""),
    }

    def worker(
        parent_url: rio.URL,
        page: rio.Page,
    ) -> None:
        cur_url = parent_url / page.page_url
        page_urls.add(cur_url)

        for child in page.children:
            worker(cur_url, child)

    for page in app.pages:
        worker(rio.URL(), page)

    # Build a XML site map
    tree = ET.Element(
        "urlset",
        xmlns="http://www.sitemaps.org/schemas/sitemap/0.9",
    )

    for relative_url in page_urls:
        full_url = base_url.with_path(relative_url.path)

        url = ET.SubElement(tree, "url")
        loc = ET.SubElement(url, "loc")
        loc.text = str(full_url)

    # Done
    return ET.tostring(tree, encoding="unicode")


@functools.lru_cache(maxsize=None)
def read_frontend_template(template_name: str) -> str:
    """
    Read a text file from the frontend directory and return its content. The
    results are cached to avoid repeated disk access.
    """
    return (common.GENERATED_DIR / template_name).read_text(encoding="utf-8")


class InitialClientMessage(uniserde.Serde):
    website_url: str
    preferred_languages: list[str]
    timezone: str
    decimal_separator: str
    thousands_separator: str
    user_settings: dict[str, Any]
    prefers_light_theme: bool

    window_width: float
    window_height: float


async def _periodically_clean_up_expired_sessions(
    app_server_ref: weakref.ReferenceType[AppServer],
):
    LOOP_INTERVAL = 60 * 15
    SESSION_LIFETIME = 60 * 60

    while True:
        await asyncio.sleep(LOOP_INTERVAL)

        app_server = app_server_ref()
        if app_server is None:
            return

        now = time.monotonic()
        cutoff = now - SESSION_LIFETIME

        for sess in app_server._active_session_tokens.values():
            if sess._last_interaction_timestamp < cutoff:
                sess.close()


class AppServer(fastapi.FastAPI):
    def __init__(
        self,
        app_: app.App,
        debug_mode: bool,
        running_in_window: bool,
        validator_factory: Callable[[rio.Session], debug.Validator] | None,
        internal_on_app_start: Callable[[], None] | None,
    ):
        super().__init__(
            title=app_.name,
            # summary=...,
            # description=...,
            openapi_url="/openapi.json" if debug_mode else None,
            docs_url="/docs" if debug_mode else None,
            redoc_url="/redoc" if debug_mode else None,
            lifespan=__class__._lifespan,
        )

        self.app = app_
        self.debug_mode = debug_mode
        self.running_in_window = running_in_window
        self.validator_factory = validator_factory
        self.internal_on_app_start = internal_on_app_start

        # Initialized lazily, when the favicon is first requested.
        self._icon_as_png_blob: bytes | None = None

        # The session tokens for all active sessions. These allow clients to
        # identify themselves, for example to reconnect in case of a lost
        # connection.
        self._active_session_tokens: dict[str, rio.Session] = {}

        # All assets that have been registered with this session. They are held
        # weakly, meaning the session will host assets for as long as their
        # corresponding Python objects are alive.
        #
        # Assets registered here are hosted under `/asset/temp-{asset_id}`. In
        # addition the server also permanently hosts other "well known" assets
        # (such as javascript dependencies) which are available under public
        # URLS at `/asset/{some-name}`.
        self._assets: weakref.WeakValueDictionary[str, assets.Asset] = (
            weakref.WeakValueDictionary()
        )

        # All pending file uploads. These are stored in memory for a limited
        # time. When a file is uploaded the corresponding future is set.
        self._pending_file_uploads: timer_dict.TimerDict[
            str, asyncio.Future[list[common.FileInfo]]
        ] = timer_dict.TimerDict(default_duration=timedelta(minutes=15))

        # FastAPI
        self.add_api_route("/robots.txt", self._serve_robots, methods=["GET"])
        self.add_api_route("/sitemap.xml", self._serve_sitemap, methods=["GET"])
        # self.add_api_route("/app.js.map", self._serve_js_map, methods=["GET"])
        # self.add_api_route("/style.css.map", self._serve_css_map, methods=["GET"])
        self.add_api_route("/rio/favicon.png", self._serve_favicon, methods=["GET"])
        self.add_api_route(
            "/rio/asset/{asset_id:path}", self._serve_asset, methods=["GET"]
        )
        self.add_api_route(
            "/rio/icon/{icon_name:path}", self._serve_icon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/upload/{upload_token}", self._serve_file_upload, methods=["PUT"]
        )
        self.add_api_websocket_route("/rio/ws", self._serve_websocket)

        # Because this is a single page application, all other routes should
        # serve the index page. The session will determine which components should
        # be shown.
        self.add_api_route(
            "/{initial_route_str:path}", self._serve_index, methods=["GET"]
        )

    async def _call_on_app_start(self) -> None:
        if self.app._on_app_start is None:
            return

        try:
            result = self.app._on_app_start(self.app)

            if inspect.isawaitable(result):
                await result

        # Display and discard exceptions
        except Exception:
            print("Exception in `on_app_start` event handler:")
            traceback.print_exc()

    async def _call_on_app_close(self) -> None:
        if self.app._on_app_close is None:
            return

        try:
            result = self.app._on_app_close(self.app)

            if inspect.isawaitable(result):
                await result

        # Display and discard exceptions
        except Exception:
            print("Exception in `_on_app_close` event handler:")
            traceback.print_exc()

    @contextlib.asynccontextmanager
    async def _lifespan(self):
        # If running as a server, periodically clean up expired sessions
        if not self.running_in_window:
            assert type(self) is AppServer  # Shut up pyright

            asyncio.create_task(
                _periodically_clean_up_expired_sessions(weakref.ref(self)),
                name="Periodic session cleanup",
            )

        # Trigger the app's startup event
        #
        # This will be done blockingly, so the user can prepare any state before
        # connections are accepted. This is also why it's called before the
        # internal event.
        await self._call_on_app_start()

        # Trigger any internal startup event
        if self.internal_on_app_start is not None:
            self.internal_on_app_start()

        try:
            yield
        finally:
            # Close all sessions
            results = await asyncio.gather(
                *(sess._close(True) for sess in self._active_session_tokens.values()),
                return_exceptions=True,
            )
            for result in results:
                if isinstance(result, BaseException):
                    print("Exception in `Session._close()`:")
                    traceback.print_exception(
                        type(result), result, result.__traceback__
                    )

            await self._call_on_app_close()

    def weakly_host_asset(self, asset: assets.HostedAsset) -> None:
        """
        Register an asset with this server. The asset will be held weakly,
        meaning the server will host assets for as long as their corresponding
        Python objects are alive.

        If another asset with the same id is already hosted, it will be
        replaced.
        """
        self._assets[asset.secret_id] = asset

    def host_asset_with_timeout(self, asset: assets.HostedAsset, timeout: float) -> URL:
        """
        Hosts an asset for a limited time. Returns the asset's url.
        """
        self.weakly_host_asset(asset)

        async def keep_alive():
            await asyncio.sleep(timeout)
            _ = asset

        asyncio.create_task(
            keep_alive(), name=f"Keep asset {asset} alive for {timeout} sec"
        )

        return asset.url

    async def _serve_index(
        self,
        request: fastapi.Request,
        initial_route_str: str,
    ) -> fastapi.responses.HTMLResponse:
        """
        Handler for serving the index HTML page via fastapi.
        """

        # Because Rio apps are single-page, this route serves as the fallback.
        # In addition to legitimate requests for HTML pages, it will also catch
        # a bunch of invalid requests to other resources. To highlight this,
        # throw a 404 if HTML is not explicitly requested.
        #
        # Currently inactive, because this caused issues behind dumb proxies
        # that don't pass on the `accept` header.

        # if not request.headers.get("accept", "").startswith("text/html"):
        #     raise fastapi.HTTPException(
        #         status_code=fastapi.status.HTTP_404_NOT_FOUND,
        #     )

        # Create a session instance to hold all of this state in an organized
        # fashion.
        #
        # The session is still missing a lot of values at this point, such as
        # `send_message` and `receive_message`. It will be finished once the
        # websocket connection is established.
        session_token = secrets.token_urlsafe()

        sess = session.Session(
            app_server_=self,
            session_token=session_token,
        )

        self._active_session_tokens[session_token] = sess

        # Add any attachments, except for user settings. These are deserialized
        # later on, once the client has sent the initial message.
        for attachment in self.app.default_attachments:
            if isinstance(attachment, user_settings_module.UserSettings):
                continue

            sess.attach(attachment)

        # Load the templates
        html = read_frontend_template("index.html")

        # Fill in any placeholders
        html = html.replace(
            "{session_token}",
            session_token,
        )

        html = html.replace(
            '"{child_attribute_names}"',
            json.dumps(
                inspection.get_child_component_containing_attribute_names_for_builtin_components()
            ),
        )

        # TODO: uvicorn has ping-pong built in, configurable via `ws_ping_interval`
        html = html.replace(
            '"{ping_pong_interval}"',
            str(self.app._ping_pong_interval.total_seconds()),
        )

        html = html.replace(
            '"{debug_mode}"',
            json.dumps(self.debug_mode),
        )

        html = html.replace(
            '"{running_in_window}"', "true" if self.running_in_window else "false"
        )

        html = html.replace("{title}", self.app.name)

        # Respond
        return fastapi.responses.HTMLResponse(html)

    async def _serve_robots(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `robots.txt` file via fastapi.
        """

        # TODO: Disallow internal API routes? Icons, assets, etc?
        request_url = URL(str(request.url))
        content = f"""
User-agent: *
Disallow:

Sitemap: {request_url.with_path("/sitemap.xml")}
        """.strip()

        return fastapi.responses.Response(
            content=content,
            media_type="text/plain",
        )

    async def _serve_sitemap(
        self, request: fastapi.Request
    ) -> fastapi.responses.Response:
        """
        Handler for serving the `sitemap.xml` file via fastapi.
        """
        return fastapi.responses.Response(
            content=_build_sitemap(rio.URL(str(request.url)), self.app),
            media_type="application/xml",
        )

    async def _serve_js_map(self) -> fastapi.responses.Response:
        """
        Handler for serving the `app.js.map` file via fastapi.
        """
        return fastapi.responses.Response(
            content=read_frontend_template("app.js.map"),
            media_type="application/json",
        )

    async def _serve_css_map(self) -> fastapi.responses.Response:
        """
        Handler for serving the `style.css.map` file via fastapi.
        """
        return fastapi.responses.Response(
            content=read_frontend_template("style.css.map"),
            media_type="application/json",
        )

    async def _serve_favicon(self) -> fastapi.responses.Response:
        """
        Handler for serving the favicon via fastapi, if one is set.
        """
        # If an icon is set, make sure a cached version exists
        if self._icon_as_png_blob is None and self.app._icon is not None:
            try:
                icon_blob, _ = await self.app._icon.try_fetch_as_blob()

                input_buffer = io.BytesIO(icon_blob)
                output_buffer = io.BytesIO()

                with Image.open(input_buffer) as image:
                    image.save(output_buffer, format="png")

            except Exception as err:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="Could not fetch the app's icon.",
                ) from err

            self._icon_as_png_blob = output_buffer.getvalue()

        # No icon set or fetching failed
        if self._icon_as_png_blob is None:
            return fastapi.responses.Response(status_code=404)

        # There is an icon, respond
        return fastapi.responses.Response(
            content=self._icon_as_png_blob,
            media_type="image/png",
        )

    async def _serve_asset(
        self,
        request: fastapi.Request,
        asset_id: str,
    ) -> fastapi.responses.Response:
        """
        Handler for serving assets via fastapi.

        Some common assets are hosted under permanent, well known URLs under the
        `/asset/{some-name}` path.

        In addition, `HostedAsset` instances are hosted under their secret id
        under the `/asset/temp-{asset_id}` path. These assets are held weakly
        by the session, meaning they will be served for as long as the
        corresponding Python object is alive.
        """

        # Special case: plotly
        #
        # The python plotly library already includes a minified version of
        # plotly.js. Rather than shipping another one, just serve the one
        # included in the library.
        if asset_id == "plotly.min.js" and plotly is not None:
            return fastapi.responses.Response(
                content=plotly.offline.get_plotlyjs(),
                media_type="text/javascript",
            )

        # Well known asset?
        if not asset_id.startswith("temp-"):
            # Construct the path to the target file
            asset_file_path = common.HOSTED_ASSETS_DIR / asset_id

            # Make sure the path is inside the hosted assets directory
            #
            # TODO: Is this safe? Would this allow the client to break out from
            # the directory using tricks such as "../", links, etc?
            asset_file_path = asset_file_path.resolve()
            if common.HOSTED_ASSETS_DIR not in asset_file_path.parents:
                logging.warning(
                    f'Client requested asset "{asset_id}" which is not located inside the hosted assets directory. Somebody might be trying to break out of the assets directory!'
                )
                return fastapi.responses.Response(status_code=404)

            return byte_serving.range_requests_response(
                request,
                file_path=asset_file_path,
            )

        # Get the asset's Python instance. The asset's id acts as a secret, so
        # no further authentication is required.
        try:
            asset = self._assets[asset_id.removeprefix("temp-")]
        except KeyError:
            return fastapi.responses.Response(status_code=404)

        # Fetch the asset's content and respond
        if isinstance(asset, assets.BytesAsset):
            return fastapi.responses.Response(
                content=asset.data,
                media_type=asset.media_type,
            )
        elif isinstance(asset, assets.PathAsset):
            return byte_serving.range_requests_response(
                request,
                file_path=asset.path,
                media_type=asset.media_type,
            )
        else:
            assert False, f"Unable to serve asset of unknown type: {asset}"

    async def _serve_icon(self, icon_name: str) -> fastapi.responses.Response:
        """
        Allows the client to request an icon by name. This is not actually the
        mechanism used by the `Icon` component, but allows JavaScript to request
        icons.
        """
        # Get the icon's SVG
        registry = components.Icon._get_registry()

        try:
            svg_source = registry.get_icon_svg(icon_name)
        except AssetError:
            return fastapi.responses.Response(status_code=404)

        # Respond
        return fastapi.responses.Response(
            content=svg_source,
            media_type="image/svg+xml",
        )

    async def _serve_file_upload(
        self,
        upload_token: str,
        file_names: list[str],
        file_types: list[str],
        file_sizes: list[str],
        # If no files are uploaded `files` isn't present in the form data at
        # all. Using a default value ensures that those requests don't fail
        # because of "missing parameters".
        #
        # Lists are mutable, make sure not to modify this value!
        file_streams: list[fastapi.UploadFile] = [],
    ):
        # Try to find the future for this token
        try:
            future = self._pending_file_uploads.pop(upload_token)
        except KeyError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_400_BAD_REQUEST,
                detail="Invalid upload token.",
            )

        # Make sure the same number of values was received for each parameter
        n_names = len(file_names)
        n_types = len(file_types)
        n_sizes = len(file_sizes)
        n_streams = len(file_streams)

        if n_names != n_types or n_names != n_sizes or n_names != n_streams:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Inconsistent number of files between the different message parts.",
            )

        # Parse the file sizes
        parsed_file_sizes: list[int] = []
        for file_size in file_sizes:
            try:
                parsed = int(file_size)
            except ValueError:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            if parsed < 0:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            parsed_file_sizes.append(parsed)

        # Complete the future
        future.set_result(
            [
                common.FileInfo(
                    name=file_names[ii],
                    size_in_bytes=parsed_file_sizes[ii],
                    media_type=file_types[ii],
                    _contents=await file_streams[ii].read(),
                )
                for ii in range(n_names)
            ]
        )

        return fastapi.responses.Response(status_code=fastapi.status.HTTP_200_OK)

    async def _serve_websocket(
        self,
        websocket: fastapi.WebSocket,
        sessionToken: str,
    ):
        """
        Handler for establishing the websocket connection and handling any
        messages.
        """
        # Blah, naming conventions
        session_token = sessionToken
        del sessionToken

        # Accept the socket
        await websocket.accept()

        # Look up the session token. If it is valid the session's duration is
        # refreshed so it doesn't expire. If the token is not valid, don't
        # accept the websocket.
        try:
            sess = self._active_session_tokens[session_token]
        except KeyError:
            # Inform the client that its session token is unknown and request a
            # refresh
            await websocket.send_json(
                {
                    "jsonrpc": "2.0",
                    "method": "invalidSessionToken",
                    "params": {},
                }
            )

            await websocket.close(
                3000,  # Custom error code
                "Invalid session token.",
            )
            return

        # Optionally create a validator
        validator_instance = (
            None if self.validator_factory is None else self.validator_factory(sess)
        )

        # Create a function for sending messages to the frontend. This function
        # will also pipe the message to the validator if one is present.
        if self.validator_factory is None:

            async def send_message(msg: uniserde.Jsonable) -> None:
                text = serialize_json(msg)
                try:
                    await websocket.send_text(text)
                except RuntimeError:  # Socket is already closed
                    pass

            async def receive_message() -> uniserde.Jsonable:
                # Refresh the session's duration
                self._active_session_tokens[session_token] = sess

                # Fetch a message
                try:
                    result = await websocket.receive_json()
                except RuntimeError:  # Socket is already closed
                    raise fastapi.WebSocketDisconnect(
                        -1
                    )  # Not clear which error code to use

                return result

            sess._send_message = send_message
            sess._receive_message = receive_message

        else:

            async def send_message(msg: uniserde.Jsonable) -> None:
                assert isinstance(validator_instance, debug.Validator)
                validator_instance.handle_outgoing_message(msg)

                text = serialize_json(msg)
                try:
                    await websocket.send_text(text)
                except RuntimeError:  # Socket is already closed
                    pass

            async def receive_message() -> uniserde.Jsonable:
                assert isinstance(validator_instance, debug.Validator)
                # Refresh the session's duration
                self._active_session_tokens[session_token] = sess

                # Fetch a message
                try:
                    msg = await websocket.receive_json()
                except RuntimeError:  # Socket is already closed
                    raise fastapi.WebSocketDisconnect(
                        -1
                    )  # Not clear which error code to use

                validator_instance.handle_incoming_message(msg)
                return msg

            sess._send_message = send_message
            sess._receive_message = receive_message

        sess._websocket = websocket
        sess._is_active_event.set()

        # Check if this is a reconnect
        try:
            if hasattr(sess, "window_width"):
                init_coro = sess._send_all_components_on_reconnect()
            else:
                await self._finish_session_initialization(
                    sess, websocket, session_token
                )

                # Trigger a refresh. This will also send the initial state to
                # the frontend.
                init_coro = sess._refresh()

            async def init():
                try:
                    await init_coro
                except Exception:
                    # If an error happens during session initialization, close
                    # the session
                    logging.exception(
                        "Unexpected exception during session initialization"
                    )
                    sess.close()

            # This is done in a task because the server is not yet running, so
            # the method would never receive a response, and thus would hang
            # indefinitely.
            asyncio.create_task(init())

            # Serve the websocket
            await sess.serve()
        except asyncio.CancelledError:
            pass

        except fastapi.WebSocketDisconnect as err:
            # If the connection was closed on purpose, close the session
            if err.code == 1001:
                sess.close()

        else:
            # I don't think this branch can even be reached, but better safe
            # than sorry, I guess?
            sess.close()
        finally:
            sess._websocket = None
            sess._is_active_event.clear()

    async def _finish_session_initialization(
        self,
        sess: session.Session,
        websocket: fastapi.WebSocket,
        session_token: str,
    ) -> None:
        # Upon connecting, the client sends an initial message containing
        # information about it. Wait for that, but with a timeout - otherwise
        # evildoers could overload the server with connections that never send
        # anything.
        initial_message_json: Jsonable = await asyncio.wait_for(
            websocket.receive_json(),
            timeout=60,
        )
        initial_message = InitialClientMessage.from_json(initial_message_json)  # type: ignore

        sess.window_width = initial_message.window_width
        sess.window_height = initial_message.window_height

        # Get locale information
        if len(initial_message.decimal_separator) != 1:
            logging.warning(
                f'Client sent invalid decimal separator "{initial_message.decimal_separator}". Using "." instead.'
            )
            sess._decimal_separator = "."
        else:
            sess._decimal_separator = initial_message.decimal_separator

        if len(initial_message.thousands_separator) > 1:
            logging.warning(
                f'Client sent invalid thousands separator "{initial_message.thousands_separator}". Using "" instead.'
            )
            sess._thousands_separator = ""
        else:
            sess._thousands_separator = initial_message.thousands_separator

        # Parse the timezone
        try:
            sess.timezone = pytz.timezone(initial_message.timezone)
        except pytz.UnknownTimeZoneError:
            logging.warning(
                f'Client sent unknown timezone "{initial_message.timezone}". Using UTC instead.'
            )
            sess.timezone = pytz.UTC

        # Set the theme according to the user's preferences
        theme = self.app._theme
        if isinstance(theme, tuple):
            if initial_message.prefers_light_theme:
                theme = theme[0]
            else:
                theme = theme[1]

        await sess._apply_theme(theme)

        # Deserialize the user settings
        await sess._load_user_settings(initial_message.user_settings)

        global_state.currently_building_component = None
        global_state.currently_building_session = sess

        sess._root_component = HighLevelRootComponent(
            self.app._build, self.app._build_connection_lost_message
        )

        global_state.currently_building_session = None

        # Trigger the `on_session_start` event.
        #
        # Since this event is often used for important initialization tasks like
        # adding attachments, actually wait for it to finish before continuing.
        await sess._call_event_handler(self.app._on_session_start, sess, refresh=False)

        # Run any page guards for the initial page
        #
        # Guards have access to the session. Thus, it should be fully
        # initialized, or at least pretend to be. Fill in any not yet final
        # values with placeholders.
        sess._base_url = (
            URL(initial_message.website_url)
            .with_path("")
            .with_query("")
            .with_fragment("")
        )
        sess._active_page_url = sess._base_url
        sess._active_page_instances = tuple()

        # Then, run the guards
        try:
            (
                active_page_instances,
                active_page_url_absolute,
            ) = routing.check_page_guards(
                sess,
                sess._base_url.join(rio.URL(initial_message.website_url)),
            )
        except routing.NavigationFailed:
            # TODO: Notify the client? Show an error?
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f'Navigation to initial page "{sess._active_page_url}" has failed.',
            ) from None

        # Is this a page, or a full URL to another site?
        try:
            common.make_url_relative(
                sess._base_url,
                active_page_url_absolute,
            )

        # This is an external URL. Navigate to it
        except ValueError:

            async def history_worker() -> None:
                await sess._evaluate_javascript(
                    f"""
                    window.location.href = {json.dumps(str(active_page_url_absolute))};
                    """
                )

            sess.create_task(history_worker(), name="navigate to external URL")

            # TODO: End the session? Abort initialization?
            return

        # Set the initial page URL. When connecting to the server, all relevant
        # page guards execute. These may change the URL of the page, so the
        # client needs to take care to update the browser's URL to match the
        # server's.
        if active_page_url_absolute != sess._active_page_url:

            async def update_url_worker():
                js_page_url = json.dumps(str(active_page_url_absolute))
                await sess._evaluate_javascript(
                    f"""
                    console.log("Updating browser URL to match the one modified by guards:", {js_page_url});
                    window.history.replaceState(null, "", {js_page_url});
                    """
                )

            sess.create_task(
                update_url_worker(),
                name="Update browser URL to match the one modified by guards",
            )

        # Update the session's active page and instances
        sess._active_page_instances = tuple(active_page_instances)
        sess._active_page_url = active_page_url_absolute



# This module silences the following warning:
#
# Task exception was never retrieved
#
# future: <Task finished name='Task-58'
# coro=<IocpProactor.accept.<locals>.accept_coro() done, defined at
# C:\Python312\Lib\asyncio\windows_events.py:558> exception=OSError(22, 'The I/O
#   operation has been aborted because of either a thread exit or an application
#     request', None, 995, None)> Traceback (most recent call last): File
# "C:\Python312\Lib\asyncio\windows_events.py", line 561, in accept_coro await
# future OSError: [WinError 995] The I/O operation has been aborted because of
# either a thread exit or an application request

import asyncio.windows_events

import introspection


def query_task_exception(task: asyncio.Task) -> None:
    try:
        task.exception()
    except:
        pass


# Since the offending function (`accept_coro`) is a nested function, the only
# solution I can think of is to monkeypatch `tasks.ensure_future`
tasks_module = asyncio.windows_events.tasks  # type: ignore


class tasks:
    @staticmethod
    def ensure_future(*args, **kwargs):
        task = asyncio.ensure_future(*args, **kwargs)
        task.add_done_callback(query_task_exception)
        return task


def accept(original_method, *args, **kwargs):
    asyncio.windows_events.tasks = tasks  # type: ignore

    try:
        return original_method(*args, **kwargs)
    finally:
        asyncio.windows_events.tasks = tasks_module  # type: ignore


introspection.wrap_method(asyncio.windows_events.IocpProactor, accept)



from .chat_message import ChatMessage
from .chat_suggestion_card import ChatSuggestionCard
from .empty_chat_placeholder import EmptyChatPlaceholder
from .generating_response_placeholder import GeneratingResponsePlaceholder



from dataclasses import KW_ONLY
from typing import final

from .fundamental_component import FundamentalComponent

__all__ = [
    "Markdown",
]


@final
class Markdown(FundamentalComponent):
    """
    # Markdown

    Displays Markdown-formatted text.

    `Markdown` displays text formatted using markdown. Markdown is a
    lightweight markup language that allows you to write text with simple
    formatting, such as **bold**, *italics*, and links.

    Markdown is a great way to write text that is both human-readable, yet
    beautifully formatted.


    ## Attributes:

    `text`: The markdown-formatted text to display.

    `default_language`: The default language to use for code blocks. If
        `None`, Rio will try to guess the language automatically. If a
        default is given, it will be used for all code blocks that don't
        specify a language explicitly.

        Inline code will always use the default language, since they are too
        short to reliably guess the language - so make sure to set a default
        language if you want your inline code to be syntax-highlighted.


    ## Example:

    This minimal example will simply display the bold text "Hello World!":

    ```python
    rio.Markdown("**Hello World!**")
    ```
    """

    text: str
    _: KW_ONLY
    default_language: str | None = None


Markdown._unique_id = "Markdown-builtin"



from __future__ import annotations

import functools
import logging
import re
from pathlib import Path
from typing import *  # type: ignore

import revel
import tomlkit
import tomlkit.exceptions
import tomlkit.items
import uniserde
from revel import *  # type: ignore

from . import rioignore

__all__ = [
    "RioProject",
]

T = TypeVar("T")


DEFAULT_FATAL = object()
DEFAULT_KEYERROR = object()


class RioProject:
    def __init__(
        self,
        file_path: Path,
        toml_dict: uniserde.JsonDoc,
        ignores: rioignore.RioIgnore,
    ) -> None:
        # Path to the `rio.toml` file. May or may not exist
        self.rio_toml_path = file_path

        # All of the data from the `rio.toml` file
        self._toml_dict = toml_dict

        # Which sections & keys have been modified and thus must be written back
        # to the `rio.toml` file
        self._dirty_keys: set[tuple[str, str]] = set()

        # Contains the parsed `.rioignore` file. When called with a path, it
        # returns True if the path should be ignored as per the `.rioignore`
        # file.
        self.ignores = ignores

        # Any rules which should be added to the `.rioignore` file. These are
        # just strings which are appended to the end of the file, one per line.
        self.rioignore_additions: list[str] = []

    def get_key(
        self,
        section_name: str,
        key_name: str,
        type: Type[T],
        default_value: Any,
    ) -> T:
        """
        Fetches the value of a key from the `rio.toml` file. If the key is
        missing, and a default value was provided it is returned instead.

        If the default value is `DEFAULT_FATAL`, an error is displayed and the
        program exits.

        If the default value is `DEFAULT_KEYERROR`, a `KeyError` is raised.
        """
        # Try to get the section
        section: dict[str, Any]

        try:
            section = self._toml_dict[section_name]  # type: ignore
        except KeyError:
            if default_value is DEFAULT_FATAL:
                fatal(
                    f"`rio.toml` is missing the `{section_name}` section",
                    status_code=1,
                )

            if default_value is DEFAULT_KEYERROR:
                raise KeyError(key_name)

            section = {}

        # Try to get the key
        try:
            value = section[key_name]

        # There is no value, use the default
        except KeyError:
            if default_value is DEFAULT_FATAL:
                fatal(
                    f"`rio.toml` is missing the `{key_name}` key. Please add it to the [[{section_name}] section.",
                    status_code=1,
                )

            if default_value is DEFAULT_KEYERROR:
                raise KeyError(key_name)

            value = default_value

        # Make sure the value is the correct type
        if not isinstance(value, type):
            fatal(
                f"`rio.toml` contains an invalid value for `{key_name}`: expected {type}, got {type(value)}",
                status_code=1,
            )

        # Done
        return value

    def set_key(self, section_name: str, key_name: str, value: Any) -> None:
        """
        Sets the value of a key in the `rio.toml` file. The value is not written
        to disk until `write()` is called.
        """
        # Get the section
        try:
            section = self._toml_dict[section_name]
        except KeyError:
            section = self._toml_dict[section_name] = {}

        # Make sure the section is indeed a section
        if not isinstance(section, dict):
            fatal(
                f"`rio.toml` contains an invalid value for `{section_name}`: this should be a section, not `{type(section)}`",
                status_code=1,
            )

        # Set the key
        section[key_name] = value
        self._dirty_keys.add((section_name, key_name))

    @property
    def project_directory(self) -> Path:
        return self.rio_toml_path.parent

    @functools.cached_property
    def module_path(self) -> Path:
        folder = self.project_directory

        # If a `src` directory exists, look there
        src_dir = folder / "src"
        if src_dir.exists():
            folder = src_dir

        # If a package (folder) exists, use that
        module_path = folder / self.app_main_module
        if module_path.exists():
            return module_path

        # Otherwise there must be a file
        return module_path.with_name(self.app_main_module + ".py")

    @property
    def app_type(self) -> Literal["app", "website"]:
        result = self.get_key("app", "app_type", str, "website")

        if result not in ("app", "website"):
            fatal(
                f"`rio.toml` contains an invalid value for `app.app_type`: It should be either `app` or `website`, not `{result}`"
            )

        return result

    @app_type.setter
    def app_type(self, value: Literal["app", "website"]) -> None:
        self.set_key("app", "app_type", value)

    @property
    def app_main_module(self) -> str:
        return self.get_key("app", "main_module", str, DEFAULT_FATAL)

    @functools.cached_property
    def deploy_name(self) -> str:
        # Is a name already stored in the `rio.toml`?
        try:
            return self.get_key("deploy", "name", str, DEFAULT_KEYERROR)
        except KeyError:
            pass

        # Ask the user for a name, and make sure it's suitable for URLs
        print(
            "What should your app be called? This name will be used as part of the URL."
        )
        print(
            'For example, if you name your app "my-app", it will be deployed at `https://rio.dev/.../my-app`.'
        )

        while True:
            name = input("App name: ")

            allowed_chars = "abcdefghijklmnopqrstuvwxyz0123456789-"
            normalized = re.sub("[^" + allowed_chars + "]", "-", name.lower())

            if name == normalized:
                break

            normalized = re.sub("-+", "-", normalized)
            print(f"`{name}` cannot be used for app names. Is `{normalized}` okay?")

            if revel.select_yes_no("", default_value=True):
                name = normalized
                break

        # Store the name
        self.set_key("deploy", "name", name)
        return name

    @staticmethod
    def _create_toml_contents_interactively(
        project_directory: Path,
    ) -> tuple[dict[str, Any], set[tuple[str, str]]]:
        """
        Interactively generates the contents for a new `rio.toml` file. The file
        isn't written anywhere, just returned as a dictionary. The second return
        value is a list of all sections and keys which were generated and should
        be considered dirty
        """
        assert project_directory.exists(), project_directory
        assert project_directory.is_dir(), project_directory

        # Find the main module
        main_module_name: str | None = None

        for subpath in project_directory.iterdir():
            if subpath.is_file():
                continue

            if subpath.is_dir() and (subpath / "__init__.py").exists():
                main_module_name = subpath.name
                break

        prompt = "What is the name of the main module?"

        if main_module_name is None:
            main_module_name = revel.input(prompt=prompt)
        else:
            main_module_name = revel.input(
                prompt="What is the name of the main module?",
                default=main_module_name,
            )

        # Build the result
        return {
            "app": {
                "app_type": "website",
                "main_module": main_module_name,
            },
        }, {
            ("app", "app_type"),
            ("app", "main_module"),
        }

    @staticmethod
    def try_locate_and_load() -> RioProject:
        """
        Best-effort attempt to locate the project directory and load the
        `rio.toml` file. This will offer to create a new project if one isn't
        found.
        """
        # Search upward until a `rio.toml` is found
        root_search_dir = Path.cwd()
        project_dir = root_search_dir

        while True:
            # Is this the project directory?
            rio_toml_path = project_dir / "rio.toml"

            if rio_toml_path.exists():
                break

            # Go up a directory, if possible
            parent_dir = project_dir.parent

            if parent_dir == project_dir:
                project_dir = root_search_dir
                break

            project_dir = parent_dir

        # Try to load the toml file
        rio_toml_path = project_dir / "rio.toml"
        dirty_keys: set[tuple[str, str]] = set()

        try:
            with rio_toml_path.open() as f:
                rio_toml_dict = tomlkit.load(f).unwrap()

        # No such file. Offer to create one
        except FileNotFoundError:
            warning(
                f"You don't appear to be inside of a [bold primary]Rio[/] project. Would you like to create one?"
            )

            if not revel.select_yes_no("", default_value=True):
                fatal(
                    f"Couldn't find `rio.toml`.",
                    status_code=1,
                )

            rio_toml_dict, dirty_keys = RioProject._create_toml_contents_interactively(
                project_dir
            )

        # Anything OS related
        except OSError as e:
            fatal(
                f"Cannot read `{rio_toml_path}`: {e}",
                status_code=1,
            )

        # Invalid syntax
        except tomlkit.exceptions.TOMLKitError as e:
            fatal(
                f"There is a problem with `rio.toml`: {e}",
                status_code=1,
            )

        # If a `.rioignore` file exists, parse it
        rioignore_path = project_dir / ".rioignore"
        ignores = rioignore.RioIgnore(base_dir=project_dir)

        try:
            with rioignore_path.open() as f:
                ignores.add_patterns_from_file(f)
        except FileNotFoundError:
            pass
        except OSError as e:
            fatal(
                f"Couldn't read `.rioignore`: {e}",
                status_code=1,
            )

        # Instantiate the project
        self = RioProject(
            file_path=rio_toml_path,
            toml_dict=rio_toml_dict,
            ignores=ignores,
        )

        self._dirty_keys = dirty_keys
        return self

    def discard_changes_and_reload(self) -> None:
        """
        Reload the `rio.toml` file, discarding any latent changes.
        """
        # Discard any latent changes
        self._dirty_keys.clear()

        # Re-load the `rio.toml` file
        with self.rio_toml_path.open() as f:
            self._toml_dict = tomlkit.load(f).unwrap()

    def __enter__(self) -> "RioProject":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # Make sure to write any changes back to the `rio.toml` file
        self.write()

    def _write_rio_toml(self) -> None:
        # Are there even any changes to write?
        if not self._dirty_keys:
            return

        logging.debug(f"Writing `{self.rio_toml_path}`")

        # Make sure the parent directory exists
        self.rio_toml_path.parent.mkdir(parents=True, exist_ok=True)

        # Fetch an up-to-date copy of the file contents, with all formatting
        # intact
        try:
            new_toml_dict = tomlkit.loads(self.rio_toml_path.read_text())

        # If it can't be read, preserve all known values
        except (OSError, tomlkit.exceptions.TOMLKitError) as e:
            new_toml_dict = tomlkit.TOMLDocument()

            self._dirty_keys.clear()
            for section_name, section in self._toml_dict.items():
                if not isinstance(section, dict):
                    continue

                for key_name, value in section.items():
                    self._dirty_keys.add((section_name, key_name))

        # Update the freshly read toml with all latent changes
        for section_name, key_name in self._dirty_keys:
            try:
                section = new_toml_dict[section_name]
            except KeyError:
                section = new_toml_dict[section_name] = tomlkit.table()

            section[key_name] = self._toml_dict[section_name][key_name]  # type: ignore

        # Write the file
        try:
            with self.rio_toml_path.open("w") as f:
                tomlkit.dump(new_toml_dict, f)

        except OSError as e:
            fatal(
                f"Couldn't write `{self.rio_toml_path}`: {e}",
                status_code=1,
            )

        # The project's keys are now clean
        self._dirty_keys.clear()

    def _write_rioignore(self) -> None:
        # Are there even any changes to write?
        if not self.rioignore_additions:
            return

        # Make sure the parent directory exists
        rioignore_path = self.project_directory / ".rioignore"
        rioignore_path.parent.mkdir(parents=True, exist_ok=True)

        # Append the lines to the file
        try:
            with rioignore_path.open("a") as f:
                f.write("\n\n")

                for line in self.rioignore_additions:
                    f.write(line)
                    f.write("\n")

        except OSError as e:
            fatal(
                f"Couldn't write `.rioignore`: {e}",
                status_code=1,
            )

    def write(self) -> None:
        """
        Write any changes back to the `rio.toml` file.
        """
        # Write the `rio.toml` file
        self._write_rio_toml()

        # Write any additions to the `.rioignore` file
        self._write_rioignore()



import rio

# <additional-imports>
from .. import components as comps
from .. import data_models

# </additional-imports>


# <component>
class CrudPage(rio.Component):
    """
    A CRUD page that allows users to create, read, update, and delete menu items.

    This component is composed of a Banner component, an ItemList component, and an ItemEditor component.

    The @rio.event.on_populate decorator is used to fetch data from a predefined data model and assign it to the
    menu_item_set attribute of the current instance. The on_press_delete_item, on_press_cancel_event, and
    on_press_save_event methods are used to handle delete, cancel, and save events, respectively. The on_press_add_new_item
    method is used to handle the add new item event. The on_press_select_menu_item method is used to handle the selection
    of a menu item.

    Attributes:
        menu_item_set: A list of menu items.
        currently_selected_menu_item: The currently selected menu item.
        banner_text: The text to be displayed in the banner.
        is_new_entry: A flag to indicate if the currently selected menu item is a new entry.
    """

    menu_item_set: list[data_models.MenuItems] = []
    currently_selected_menu_item: data_models.MenuItems | None = None
    banner_text: str = ""
    is_new_entry: bool = False

    @rio.event.on_populate
    def on_populate(self) -> None:
        """
        Event handler that is called when the component is populated.

        Fetches data from a predefined data model and assigns it to the menu_item_set
        attribute of the current instance.
        """
        self.menu_item_set = data_models.MENUITEMSET

    async def on_press_delete_item(self, idx: int) -> None:
        """
        Perform actions when the "Delete" button is pressed.

        Args:
            idx: The index of the item to be deleted.
        """
        self.menu_item_set.pop(idx)
        self.banner_text = "Item was deleted"
        self.currently_selected_menu_item = None

    async def on_press_cancel_event(self) -> None:
        """
        Perform actions when the "Cancel" button is pressed.
        """
        self.currently_selected_menu_item = None
        self.banner_text = ""

    def on_press_save_event(self) -> None:
        """
        Performs actions when the "Save" button is pressed.

        This method appends the currently selected menu item to the menu item set if it is a new entry,
        or updates the menu item set if it is an existing entry. It also updates the banner text and sets
        the is_new_entry flag to False.
        """
        assert self.currently_selected_menu_item is not None
        if self.is_new_entry:
            self.menu_item_set.append(self.currently_selected_menu_item)
            self.banner_text = "Item was added"
            self.is_new_entry = False
            self.currently_selected_menu_item = None
        else:
            self.banner_text = "Item was updated"

    async def on_press_add_new_item(self) -> None:
        """
        Perform actions when the "Add New" ListItem is pressed.

        This method sets the currently selected menu item to a new empty instance of models.MenuItems,
        clears the banner text, and sets the is_new_entry flag to True.
        """
        self.currently_selected_menu_item = data_models.MenuItems.new_empty()
        self.banner_text = ""
        self.is_new_entry = True

    async def on_press_select_menu_item(
        self, selected_menu_item: data_models.MenuItems
    ) -> None:
        """
        Perform actions when a menu item is selected.

        This method sets the currently selected menu item to the selected menu item,
        which is passed as an argument.

        Args:
            selected_menu_item: The selected menu item.
        """
        self.currently_selected_menu_item = selected_menu_item

    def build(self) -> rio.Component:
        """
        Builds the component to be rendered.

        If there is no currently selected menu item, only the Banner and ItemList component is returned.

        Otherwise, if there is a currently selected menu item, both the Banner and ItemList component
        and the ItemEditor component are returned.

        Returns:
            rio.Component: The components to be rendered.
            See the approx. layout below:

        ############### Column ###############
        #  Banner                            #
        #  +-------------------------------  #
        #  | ItemList                     |  #
        #  |                              |  #
        #  +-------------------------------  #
        ######################################

        or

        ############### Column ###############
        #  Banner                            #
        #  +-------------------------------  #
        #  | ItemList        | ItemEditor |  #
        #  |                 |            |  #
        #  +-------------------------------  #
        ######################################

        """

        if self.currently_selected_menu_item is None:
            return rio.Column(
                rio.Banner(self.banner_text, style="danger"),
                comps.ItemList(
                    menu_item_set=self.menu_item_set,
                    on_add_new_item_event=self.on_press_add_new_item,
                    on_delete_item_event=self.on_press_delete_item,
                    on_select_item_event=self.on_press_select_menu_item,
                    align_y=0,
                ),
                align_y=0,
                margin=3,
            )
        else:
            return rio.Column(
                rio.Banner(self.banner_text, style="danger"),
                rio.Row(
                    comps.ItemList(
                        menu_item_set=self.menu_item_set,
                        on_add_new_item_event=self.on_press_add_new_item,
                        on_delete_item_event=self.on_press_delete_item,
                        on_select_item_event=self.on_press_select_menu_item,
                        align_y=0,
                    ),
                    comps.ItemEditor(
                        self.currently_selected_menu_item,
                        new_entry=self.is_new_entry,
                        on_cancel_event=self.on_press_cancel_event,
                        on_save_event=self.on_press_save_event,
                    ),
                    spacing=1,
                    proportions=(2, 1),
                ),
                spacing=1,
                align_y=0,
                margin=3,
            )


# </component>



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeVar, final

from uniserde import JsonDoc

import rio

from .. import icon_registry
from .fundamental_component import FundamentalComponent

__all__ = [
    "SwitcherBarChangeEvent",
    "SwitcherBar",
]

T = TypeVar("T")


@final
@dataclass
class SwitcherBarChangeEvent(Generic[T]):
    value: T | None


@final
class SwitcherBar(FundamentalComponent, Generic[T]):
    """
    # SwitcherBar

    A bar of options which can be switched between.

    A `SwitcherBar` is a bar of options which can be switched between. It is
    commonly used to switch between different views or modes.


    ## Attributes:

    `values`: The list of values which can be selected.

    `names`: The list of names to display for each value. If `None`, the
        string representation of each value is used.

    `icons`: The list of icons to display for each value.

    `color`: The color of the switcher bar.

    `orientation`: The orientation of the switcher bar.

    `spacing`: The spacing between options.

    `selected_value`: The currently selected value.

    `allow_none`: Whether the switcher bar can have no value selected.

    `on_change`: Triggered whenever the selected value changes.


    ## Example:

    A simple switcher bar with three options:

    ```python
    rio.SwitcherBar(
        values=[1, 2, 3],
        names=["A-SwitchName", "B-SwitchName", "C-SwitchName"],
        selected_value=1,
        on_change=lambda event: print(event.value),
    )
    ```

    You can use a `SwitcherBar` to create your own custom Navigation Bar. use
    the on_page_change event to trigger a refresh of the `SwitcherBar` when the
    page changes. Use the page_url defined in your rio.App and rio.Page
    instances to navigate to the selected page. Here is an example of a custom
    `NavigationBar` component:

    ```python
    class NavigationBar(rio.Component):
        @rio.event.on_page_change
        async def _on_page_change(self) -> None:
            await self.force_refresh()

        def _on_change(self, event: rio.SwitcherBarChangeEvent) -> None:
            # navigate to the selected page
            assert isinstance(event.value, str)
            self.session.navigate_to(event.value)

        def build(self) -> rio.Component:
            return rio.Card(
                rio.Column(
                    rio.Row(
                        rio.Image(
                            self.session.assets / "example.png",
                            width=10,
                            height=2,
                        ),
                        rio.Spacer(),
                        rio.SwitcherBar(
                            # values are the page_urls, which are the same as the page_url
                            # of the rio.Page instances, e.g.: rio.Page(page_url="first-page", ...)
                            # rio.App -> rio.Page(page_url="first-page", ...) -> "first-page"
                            values=["/", "first-page", "second-page"],
                            names=["Home", "First Page", "Second Page"],
                            selected_value=self.session.active_page_instances[0].page_url,
                            align_y=0.5,
                            color="primary",
                            on_change=self._on_change,
                        ),
                        margin=1,
                        width="grow",
                    ),
                    rio.Rectangle(
                        width="grow",
                        height=0.2,
                        fill=rio.Color.from_hex("ffdc00"),
                        margin_bottom=2,
                    ),
                ),
                width="grow",
                corner_radius=0,
                color="background",
            )
    ```
    """

    names: list[str]
    values: list[T]
    icon_svg_sources: list[str | None]
    color: rio.ColorSet
    orientation: Literal["horizontal", "vertical"]
    spacing: float
    selected_value: T | None
    allow_none: bool
    on_change: rio.EventHandler[SwitcherBarChangeEvent[T]]

    def __init__(
        self,
        values: list[T],
        *,
        names: list[str] | None = None,
        icons: Sequence[str | None] | None = None,
        color: rio.ColorSet = "keep",
        orientation: Literal["horizontal", "vertical"] = "horizontal",
        spacing: float = 1.0,
        allow_none: bool = False,
        selected_value: T | None = None,
        on_change: rio.EventHandler[SwitcherBarChangeEvent[T]] = None,
        key: str | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        width: float | Literal["natural", "grow"] = "natural",
        height: float | Literal["natural", "grow"] = "natural",
        align_x: float | None = None,
        align_y: float | None = None,
    ):
        if not values:
            raise ValueError("`SwitcherBar` must have at least one option.")

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.values = values
        self.color = color
        self.orientation = orientation
        self.spacing = spacing
        self.allow_none = allow_none
        self.on_change = on_change

        # Names default to the string representation of the values
        if names is None:
            self.names = [str(value) for value in values]
        else:
            if len(names) != len(values):
                raise ValueError("`names` must be the same length as `values`.")

            self.names = names

        # Icons default to `None`. Also, fetch their SVG sources so any errors
        # are raised now, rather than later.
        if icons is None:
            self.icon_svg_sources = [None] * len(values)
        else:
            if len(icons) != len(values):
                raise ValueError("`icons` must be the same length as `values`.")

            registry = icon_registry.IconRegistry.get_singleton()
            self.icon_svg_sources = [
                None if icon is None else registry.get_icon_svg(icon) for icon in icons
            ]

        self.selected_value = selected_value

    def __post_init__(self) -> None:
        # Make sure a value is selected, if needed
        if self.selected_value is None and not self.allow_none:
            self.selected_value = self.values[0]

    def _fetch_selected_name(self) -> str | None:
        # None is fine
        if self.selected_value is None:
            assert self.allow_none
            return None

        # The frontend works with names, not values. Get the corresponding name.

        # Avoid hammering a potential state binding
        selected_value = self.selected_value

        # Fetch the name
        for name, value in zip(self.names, self.values):
            if value == selected_value:
                return name
        else:
            # If nothing matches, just select the first option
            return self.names[0]

    def _custom_serialize(self) -> JsonDoc:
        result = {
            "optionNames": self.names,
            "optionIcons": self.icon_svg_sources,
            "selectedName": self._fetch_selected_name(),
            "color": self.session.theme._serialize_colorset(self.color),
        }

        return result

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        # The frontend works with names, not values. Get the corresponding
        # value.
        selected_name = msg["name"]

        if selected_name is None:
            # Invalid names may be sent due to lag between the frontend and
            # backend. Ignore them.
            if not self.allow_none:
                return

            selected_value = None

        else:
            for name, value in zip(self.names, self.values):
                if name == selected_name:
                    selected_value = value
                    break
            else:
                # Invalid names may be sent due to lag between the frontend and
                # backend. Ignore them.
                return

        # TEMP, for debugging the switcher bar JS code
        # self._apply_delta_state_from_frontend({'selected_value': selected_value})
        self.selected_value = selected_value

        # Trigger the event
        await self.call_event_handler(
            self.on_change, SwitcherBarChangeEvent(self.selected_value)
        )

        # Refresh the session
        await self.session._refresh()


SwitcherBar._unique_id = "SwitcherBar-builtin"
